from newclid.ggb.elements_relationships import (
    BisectorLineRelationship,
    BisectorPointRelationship,
    ElementsRelationships,
    OrthoRelationship,
    ParaRelationship,
    TangentRelationship,
)
from newclid.ggb.read_commands import GGBCommand, GGBReadCommandType
from newclid.ggb.read_elements import GGBConic, GGBLine, GGBPoint


def relationships_from_commands(
    commands: list[GGBCommand],
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> list[ElementsRelationships]:
    relationships: list[ElementsRelationships] = []
    for command in commands:
        relationships.extend(
            _relationships_from_command(command, points, lines, conics)
        )
    return relationships


def _relationships_from_command(
    command: GGBCommand,
    points: dict[str, GGBPoint],
    lines: dict[str, GGBLine],
    conics: dict[str, GGBConic],
) -> list[ElementsRelationships]:
    match command.construction_type:
        case GGBReadCommandType.OrthogonalLine:
            return [
                OrthoRelationship(
                    line=lines[command.perpendicular_line],
                    orthogonal_to=lines[command.line],
                )
            ]
        case GGBReadCommandType.ParallelLine:
            return [
                ParaRelationship(
                    line=lines[command.base_line],
                    parallel_to=lines[command.parallel_line],
                )
            ]
        case GGBReadCommandType.AngularBisector:
            return [
                BisectorLineRelationship(
                    line=lines[command.line_bisector],
                    bisector_1=lines[command.line_1],
                    bisector_2=lines[command.line_2],
                )
            ]
        case GGBReadCommandType.AngularBisectorPoints:
            return [
                BisectorPointRelationship(
                    line=lines[command.line_bisector],
                    angle_a=points[command.point_a],
                    angle_vertex=points[command.corner],
                    angle_b=points[command.point_b],
                )
            ]
        case GGBReadCommandType.TangentOnCircle:
            return [
                TangentRelationship(
                    line=lines[command.tangent_line],
                    tangent_to=conics[command.tangent_object],
                )
            ]
        case GGBReadCommandType.TangentOutsideCircle:
            return [
                TangentRelationship(
                    line=lines[command.tangent_line1],
                    tangent_to=conics[command.tangent_object],
                ),
                TangentRelationship(
                    line=lines[command.tangent_line2],
                    tangent_to=conics[command.tangent_object],
                ),
            ]
        case _:
            return []



from enum import Enum


class JustificationType(str, Enum):
    ASSUMPTION = "Assumption"
    NUMERICAL_CHECK = "Numerical Check"
    RULE_APPLICATION = "Rule Application"
    AR_DEDUCTION = "AR Deduction"
    CIRCLE_MERGE = "Circle Merge"
    LINE_MERGE = "Line Merge"
    DIRECT_CONSEQUENCE = "Direct Consequence"
    REFLEXIVITY = "Reflexivity"


class JustificationError(Exception):
    """Exception raised when a justification is not possible."""



from abc import ABC, abstractmethod

import numpy
from pydantic import BaseModel

from newclid.heuristics.geometric_objects import (
    AngleHeuristic,
    CircleHeuristic,
    LineHeuristic,
)
from newclid.jgex.clause import JGEXClause
from newclid.predicate_types import PredicateArgument
from newclid.problem import PredicateConstruction


class HeuristicSetup(BaseModel):
    """Problem given to a heuristic to deduce new JGEX clauses."""

    points: tuple[PredicateArgument, ...]
    free_points: tuple[PredicateArgument, ...]
    lines: tuple[LineHeuristic, ...]
    circles: tuple[CircleHeuristic, ...]
    angles: tuple[AngleHeuristic, ...]
    goals: tuple[PredicateConstruction, ...]


class Heuristic(ABC):
    @abstractmethod
    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        """Find new clauses to add to the problem by applying the heuristic."""



"""Define the problem setup taken as input by Newclid."""

from __future__ import annotations

from typing import Self

from pydantic import BaseModel, field_validator, model_validator

from newclid.predicate_types import PredicateArgument
from newclid.predicates import (
    Predicate,
    predicate_class_from_type,
    predicate_from_construction,
)
from newclid.predicates._index import PredicateType
from newclid.symbols.points_registry import Point, PointsRegisty
from newclid.tools import point_construction_tuple

SEPARATOR = " "


class PredicateConstruction(BaseModel):
    string: str
    """String representation of the construction."""

    @property
    def predicate_type(self) -> PredicateType:
        return PredicateType(self.string.split(SEPARATOR)[0])

    @property
    def args(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(a) for a in self.string.split(SEPARATOR)[1:])

    @model_validator(mode="after")
    def canonicalize(self) -> Self:
        predicate_class = predicate_class_from_type(self.predicate_type)
        parsed_args = predicate_class.preparse(
            tuple(PredicateArgument(a) for a in self.args)
        )
        if parsed_args is None:
            raise ValueError(f"Invalid construction: {self}")
        self.string = SEPARATOR.join(
            (self.predicate_type.value, *tuple(str(a) for a in parsed_args))
        )
        return self

    @classmethod
    def from_tuple(cls, construction_tuple: tuple[str, ...]) -> PredicateConstruction:
        return cls(string=SEPARATOR.join(construction_tuple))

    @classmethod
    def from_str(cls, construction_str: str) -> PredicateConstruction:
        return cls(string=construction_str)

    @classmethod
    def from_predicate_type_and_args(
        cls, predicate_type: PredicateType, args: tuple[PredicateArgument, ...]
    ) -> PredicateConstruction:
        return cls.from_tuple((predicate_type.value, *args))

    def __str__(self) -> str:
        return self.string

    def __repr__(self) -> str:
        return self.string

    def __hash__(self) -> int:
        return hash(self.string)


def predicate_from_str(predicate_str: str, points: PointsRegisty) -> Predicate:
    construction = PredicateConstruction.from_str(predicate_str)
    predicate = predicate_from_construction(construction, points)
    if predicate is None:
        raise ValueError(f"Invalid predicate string: {predicate_str}")
    return predicate


def predicate_to_construction(predicate: Predicate) -> PredicateConstruction:
    return PredicateConstruction.from_predicate_type_and_args(
        predicate_type=predicate.predicate_type, args=predicate.to_tokens()
    )


def is_numerical_argument(arg: str) -> bool:
    return not str.isalpha(arg[0])


def predicate_points(predicate: Predicate) -> set[PredicateArgument]:
    construction = predicate_to_construction(predicate)
    return {pname for pname in construction.args if not is_numerical_argument(pname)}


def rename_predicate_construction(
    predicate_construction: PredicateConstruction,
    mapping: dict[PredicateArgument, PredicateArgument],
) -> PredicateConstruction:
    """Rename the points in a predicate construction."""
    return PredicateConstruction.from_tuple(
        (
            predicate_construction.predicate_type.value,
            *tuple(mapping.get(arg, arg) for arg in predicate_construction.args),
        )
    )


class ProblemSetup(BaseModel):
    name: str = "problem"
    points: tuple[Point, ...]
    assumptions: tuple[PredicateConstruction, ...]
    goals: tuple[PredicateConstruction, ...]

    @field_validator("points")
    def check_points_are_unique(cls, v: tuple[Point, ...]) -> tuple[Point, ...]:
        if len(v) != len(set(p.name for p in v)):
            raise ValueError("Points must be unique")
        return v

    @model_validator(mode="after")
    def check_assumptions_and_goals_refer_to_existing_points(self) -> Self:
        point_names = {p.name for p in self.points}
        for construction in self.assumptions + self.goals:
            for arg in construction.args:
                if str.isalpha(arg[0]) and arg not in point_names:
                    raise ValueError(
                        f"Construction {construction} refers to non-existing point {arg}"
                    )
        self.points = tuple(
            sorted(self.points, key=lambda p: point_construction_tuple(p.name))
        )
        self.assumptions = tuple(sorted(self.assumptions, key=str))
        self.goals = tuple(sorted(self.goals, key=str))
        return self

    def with_new(
        self,
        *,
        new_points: tuple[Point, ...] = (),
        new_assumptions: tuple[PredicateConstruction, ...] = (),
        new_goals: tuple[PredicateConstruction, ...] = (),
    ) -> ProblemSetup:
        return ProblemSetup(
            points=self.points + new_points,
            assumptions=self.assumptions + new_assumptions,
            goals=self.goals + new_goals,
        )

    def pretty_str(self) -> str:
        result = f"Built the problem {self.name} with assumptions:\n"
        for assumption in self.assumptions:
            result += f"{assumption}\n"
        result += "\nand goals:\n"
        for goal in self.goals:
            result += f"{goal}\n"

        result += "\nUsing points:\n"
        for point in self.points:
            result += f"{point.name} ({point.num.x}, {point.num.y})\n"
        return result


def nc_problem_is_valid(problem: ProblemSetup) -> bool:
    """Check if a problem is valid."""
    points_registry = PointsRegisty()
    for point in problem.points:
        points_registry.add_point(point)
    for predicate_construction in problem.assumptions + problem.goals:
        if not _predicate_construction_is_valid(
            predicate_construction, points_registry
        ):
            return False
    return True


def _predicate_construction_is_valid(
    predicate_construction: PredicateConstruction,
    points_registry: PointsRegisty,
) -> bool:
    """Check if a predicate is valid."""
    predicate = predicate_from_construction(predicate_construction, points_registry)
    return predicate is not None and predicate.check_numerical()


def rename_points_in_nc_problem(
    nc_problem: ProblemSetup, mapping: dict[PredicateArgument, PredicateArgument]
) -> ProblemSetup:
    """Rename the points in the problem and the apply the change to the assumptions and goals.

    Points not present in the mapping will be ignored.
    """
    existing_points = {p.name for p in nc_problem.points}
    for old_name in mapping.keys():
        if old_name not in existing_points:
            raise ValueError(f"Point {old_name} does not exist and yet is mapped.")

    renamed_points = tuple(
        Point(name=mapping.get(p.name, p.name), num=p.num) for p in nc_problem.points
    )
    renamed_assumptions = tuple(
        rename_predicate_construction(a, mapping) for a in nc_problem.assumptions
    )
    renamed_goals = tuple(
        rename_predicate_construction(g, mapping) for g in nc_problem.goals
    )
    return ProblemSetup(
        points=renamed_points, assumptions=renamed_assumptions, goals=renamed_goals
    )


def filter_points_from_nc_problem(
    nc_problem: ProblemSetup, points_to_keep: list[PredicateArgument]
) -> ProblemSetup:
    """Filter the points in the problem and remove the assumptions and goals that refer to unkept points."""
    if len(points_to_keep) == 0:
        raise ValueError(
            "There is no point to keep, create a new problem instead if this is expected."
        )

    points_kept_names = set(
        p.name for p in nc_problem.points if p.name in points_to_keep
    )
    if len(points_kept_names) != len(points_to_keep):
        raise ValueError("Some points to keep do not exist in the problem.")

    points_kept = tuple(p for p in nc_problem.points if p.name in points_kept_names)
    assumptions_kept = tuple(
        a
        for a in nc_problem.assumptions
        if all(arg in points_kept_names for arg in a.args)
    )
    goals_kept = tuple(
        g for g in nc_problem.goals if all(arg in points_kept_names for arg in g.args)
    )
    return ProblemSetup(
        points=points_kept, assumptions=assumptions_kept, goals=goals_kept
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from fractions import Fraction
from typing import TYPE_CHECKING, Iterator, Literal, Self

from pydantic import BaseModel, model_validator

from newclid.deductors import ARReason
from newclid.justifications._index import JustificationType
from newclid.predicates import Predicate
from newclid.symbols.symbols_registry import SymbolsRegistry

if TYPE_CHECKING:
    from newclid.justifications.justification import Justification


class ARCoefficient(BaseModel):
    """A premise of an AR deduction."""

    coeff: Fraction
    """The coefficient with which the predicate is multiplied to get the final equation for the wanted predicate."""
    lhs_terms: dict[str, Fraction]
    """The left-hand side terms of the equation in the table."""

    def __hash__(self) -> int:
        return hash((self.coeff, tuple(sorted(self.lhs_terms.items()))))


class ARPremise(BaseModel):
    """A premise of an AR deduction."""

    predicate: Predicate
    """The predicate construction of the premise."""
    coefficient: ARCoefficient
    """The coefficient of the premise."""

    def __hash__(self) -> int:
        return hash((self.predicate, self.coefficient))


class ARDeduction(BaseModel):
    predicate: Predicate
    """The predicate justified by the AR deduction."""
    ar_reason: ARReason
    """The reason for the AR deduction."""
    ar_premises: tuple[ARPremise, ...] | None = None
    """The AR premises of the AR deduction."""

    dependency_type: Literal[JustificationType.AR_DEDUCTION] = (
        JustificationType.AR_DEDUCTION
    )

    @model_validator(mode="after")
    def canonicalize(self) -> Self:
        if self.ar_premises is not None:
            self.ar_premises = tuple(
                sorted(set(self.ar_premises), key=lambda x: repr(x))
            )
        return self

    def __str__(self) -> str:
        premises_txt = ""
        if self.ar_premises is not None:
            premises_txt = (
                ", ".join(str(premise.predicate) for premise in self.ar_premises) + " "
            )
        return f"{premises_txt}=({self.ar_reason.value})> {self.predicate}"

    def __hash__(self) -> int:
        return hash(self.predicate)


class Deductor(ABC):
    @abstractmethod
    def deduce(self, symbols_registry: SymbolsRegistry) -> Iterator[Justification]:
        """Deduce dependencies o new predicates from the current dependency graph."""

    @abstractmethod
    def add_dependency(
        self, dependency: Justification, symbols_registry: SymbolsRegistry
    ) -> None:
        """Add a predicate to the deductor."""

    @abstractmethod
    def check_predicate(
        self, predicate: Predicate, symbols_registry: SymbolsRegistry
    ) -> bool:
        """Check if a predicate is valid for the deductor."""

    @abstractmethod
    def justify_predicate(
        self, predicate: Predicate, symbols_registry: SymbolsRegistry
    ) -> Justification | None:
        """Justify a predicate with a dependency."""


def add_to_deductors(
    dependency: Justification,
    deductors: list[Deductor],
    symbols_registry: SymbolsRegistry,
) -> None:
    for deductor in deductors:
        deductor.add_dependency(dependency, symbols_registry)


def check_from_deductors(
    predicate: Predicate, deductors: list[Deductor], symbols_registry: SymbolsRegistry
) -> bool:
    for deductor in deductors:
        if deductor.check_predicate(predicate, symbols_registry):
            return True
    return False


def justify_from_deductors(
    predicate: Predicate, deductors: list[Deductor], symbols_registry: SymbolsRegistry
) -> Justification | None:
    for deductor in deductors:
        why_dep = deductor.justify_predicate(predicate, symbols_registry)
        if why_dep is not None:
            return why_dep
    return None



"""Numerical representation of geometric objects in Newclid."""

ATOM = 1e-7
REL_TOL = 0.001


def close_enough(a: float, b: float) -> bool:
    return abs(a - b) < 4 * ATOM or abs(a - b) / max(abs(a), abs(b)) < REL_TOL


def nearly_zero(a: float) -> bool:
    return abs(a) < 2 * ATOM


def sign(a: float) -> int:
    return 0 if nearly_zero(a) else (1 if a > 0 else -1)



class JGEXConstructionError(Exception):
    """Error during JGEX construction."""


class PointTooCloseError(JGEXConstructionError):
    """Trying to create a point that is too close to an existing point."""


class PointTooFarError(JGEXConstructionError):
    """Trying to create a point that is too far from existing points."""


class InvalidIntersectError(JGEXConstructionError):
    """Unexpected number of intersections points."""



from typing import cast

from newclid.ggb.read_elements import GGBConic, GGBLine
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction


def elements_to_newclid_assumptions(
    lines: dict[str, GGBLine], conics: dict[str, GGBConic]
) -> list[PredicateConstruction]:
    assumptions: list[PredicateConstruction] = []
    for line in lines.values():
        assumptions.extend(_line_to_newclid_assumptions(line))
    for conic in conics.values():
        assumptions.extend(_conic_to_newclid_assumptions(conic))
    return assumptions


def _line_to_newclid_assumptions(line: GGBLine) -> list[PredicateConstruction]:
    assumptions: list[PredicateConstruction] = []
    if len(line.points) <= 1:
        return assumptions
    for i in range(len(line.points) - 2):
        coll_points = line.points[i : i + 3]
        new_predicate = PredicateConstruction.from_predicate_type_and_args(
            PredicateType.COLLINEAR, cast(tuple[PredicateArgument, ...], coll_points)
        )
        assumptions.append(new_predicate)
    return assumptions


def _conic_to_newclid_assumptions(conic: GGBConic) -> list[PredicateConstruction]:
    cong_predicates: list[PredicateConstruction] = []
    center_predicates: list[PredicateConstruction] = []
    cyclic_predicates: list[PredicateConstruction] = []
    if len(conic.points) <= 2:
        if conic.radius is None or conic.center is None:
            return []
        else:
            radius_points = [conic.radius[0], conic.radius[1]]
            for point in conic.points:
                args = [point, conic.center, *radius_points]
                if len(set(args)) > 2:
                    new_predicate = PredicateConstruction.from_predicate_type_and_args(
                        PredicateType.CONGRUENT,
                        cast(tuple[PredicateArgument, ...], args),
                    )
                    cong_predicates.append(new_predicate)

            return cong_predicates

    elif conic.center is not None:
        for i in range(0, len(conic.points) - 2):
            args = [
                conic.center,
                conic.points[i],
                conic.points[i + 1],
                conic.points[i + 2],
            ]
            center_predicates.append(
                PredicateConstruction.from_predicate_type_and_args(
                    PredicateType.CIRCUMCENTER,
                    cast(tuple[PredicateArgument, ...], args),
                )
            )
        if conic.radius is not None:
            radius_points = [conic.radius[0], conic.radius[1]]
            for point in conic.points:
                args = [point, conic.center, *radius_points]
                if set(args[:2]) == set(args[2:]):
                    # Avoid trivial cong A B B A
                    continue
                new_predicate = PredicateConstruction.from_predicate_type_and_args(
                    PredicateType.CONGRUENT,
                    cast(tuple[PredicateArgument, ...], args),
                )
                cong_predicates.append(new_predicate)

        if len(conic.points) > 3:
            for i in range(0, len(conic.points) - 3):
                args = [
                    conic.points[i],
                    conic.points[i + 1],
                    conic.points[i + 2],
                    conic.points[i + 3],
                ]
                cyclic_predicates.append(
                    PredicateConstruction.from_predicate_type_and_args(
                        PredicateType.CYCLIC,
                        cast(tuple[PredicateArgument, ...], args),
                    )
                )
    return cong_predicates + center_predicates + cyclic_predicates



import itertools
import logging
from typing import Literal

import numpy
from pydantic import BaseModel

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic, HeuristicSetup
from newclid.heuristics.alphabet import get_available_from_alphabet
from newclid.jgex.clause import JGEXClause

LOGGER = logging.getLogger(__name__)


class FootHeuristicConfig(BaseModel):
    heuristic_name: Literal[HeuristicName.FOOT] = HeuristicName.FOOT


class FootHeuristic(Heuristic):
    """Add foot points on lines."""

    def new_clauses(
        self,
        setup: HeuristicSetup,
        max_new_points: int,
        rng: numpy.random.Generator,
    ) -> list[JGEXClause]:
        points = list(setup.points)
        pairs_of_points = list(itertools.combinations(points, 2))
        unflattened_triplets = itertools.product(points, pairs_of_points)
        triplets_of_points = [(a, b, c) for a, (b, c) in unflattened_triplets]
        new_clauses: list[JGEXClause] = []

        for _ in range(max_new_points):
            if not triplets_of_points:
                break

            triplet_index = rng.choice(len(triplets_of_points))
            p3, p1, p2 = triplets_of_points.pop(triplet_index)

            new_point = get_available_from_alphabet(points)
            if new_point is None:
                LOGGER.warning(
                    "No available points left in the alphabet, skipping foot point addition."
                )
                break

            points.append(new_point)
            new_clause_string = f"{new_point} = foot {new_point} {p3} {p1} {p2}"
            new_clause = JGEXClause.from_str(new_clause_string)[0]
            new_clauses.append(new_clause)

        return new_clauses



"""Implements objects to represent problems, theorems, proofs, traceback."""

from __future__ import annotations

import string
from pathlib import Path
from typing import Literal

from pydantic import BaseModel

from newclid.jgex.clause import JGEXClause
from newclid.predicate_types import PredicateArgument
from newclid.problem import PredicateConstruction, rename_predicate_construction
from newclid.tools import atomize, points_by_construction_order, reshape

ALPHABET: list[str] = (
    list(string.ascii_lowercase)
    + [c + "1" for c in string.ascii_lowercase]
    + [c + "2" for c in string.ascii_lowercase]
    + [c + "3" for c in string.ascii_lowercase]
    + [c + "4" for c in string.ascii_lowercase]
    + [c + "5" for c in string.ascii_lowercase]
    + [c + "6" for c in string.ascii_lowercase]
)

CLAUSES_SEPARATOR = "; "
AUX_SEPARATOR = " | "
GOAL_SEPARATOR = " ? "
MULTI_GOALS_SEPARATOR = "; "


class JGEXFormulation(BaseModel):
    """Describe one problem to solve."""

    name: str = "ProblemJGEX"
    setup_clauses: tuple[JGEXClause, ...]
    auxiliary_clauses: tuple[JGEXClause, ...] = ()
    goals: tuple[PredicateConstruction, ...]

    formulation_type: Literal["jgex"] = "jgex"

    @property
    def clauses(self) -> tuple[JGEXClause, ...]:
        return self.setup_clauses + self.auxiliary_clauses

    @classmethod
    def from_text(cls, text: str) -> JGEXFormulation:
        """Load a problem from a string."""
        name = ""
        if "\n" in text:
            name, text = text.split("\n")

        if GOAL_SEPARATOR in text:
            constructions_str, goals_str = atomize(text, GOAL_SEPARATOR)
        else:
            constructions_str, goals_str = text, ""
        goals_strs = atomize(goals_str, MULTI_GOALS_SEPARATOR)
        goals: tuple[PredicateConstruction, ...] = tuple()
        if len(goals_strs) > 0:
            goals = tuple(PredicateConstruction.from_str(g) for g in goals_strs if g)

        if AUX_SEPARATOR in constructions_str:
            setup_str, aux_str = atomize(constructions_str, AUX_SEPARATOR)
            auxiliary_clauses = JGEXClause.from_str(aux_str)
        else:
            setup_str, aux_str = constructions_str, ""
            auxiliary_clauses = ()

        setup_clauses = JGEXClause.from_str(setup_str)
        return cls(
            name=name,
            setup_clauses=setup_clauses,
            auxiliary_clauses=auxiliary_clauses,
            goals=goals,
        )

    @property
    def points(self) -> tuple[PredicateArgument, ...]:
        s: set[PredicateArgument] = set()
        for construction in self.clauses:
            s.update(construction.points)
        return tuple(points_by_construction_order(s))

    def renamed(
        self, mapping: dict[PredicateArgument, PredicateArgument]
    ) -> JGEXFormulation:
        renamed_setup_clauses = tuple(
            clause.renamed(mapping) for clause in self.setup_clauses
        )
        renamed_aux_clauses = tuple(
            clause.renamed(mapping) for clause in self.auxiliary_clauses
        )
        renamed_goals = tuple(
            rename_predicate_construction(goal, mapping) for goal in self.goals
        )
        return JGEXFormulation(
            name=self.name,
            setup_clauses=renamed_setup_clauses,
            auxiliary_clauses=renamed_aux_clauses,
            goals=renamed_goals,
        )

    def __str__(self) -> str:
        setup_str = CLAUSES_SEPARATOR.join(str(c) for c in self.setup_clauses)
        goals_str = MULTI_GOALS_SEPARATOR.join(str(goal) for goal in self.goals)
        aux_str = ""
        if self.auxiliary_clauses:
            aux_str = AUX_SEPARATOR + CLAUSES_SEPARATOR.join(
                str(c) for c in self.auxiliary_clauses
            )
        return setup_str + aux_str + GOAL_SEPARATOR + goals_str

    def __hash__(self) -> int:
        return hash(str(self))


def jgex_formulation_from_txt_file(fname: Path) -> dict[str, JGEXFormulation]:
    with open(fname, "r") as f:
        lines = f.read().split("\n")

    lines = [line for line in lines if line and not line.startswith("#")]
    problems = [
        JGEXFormulation.from_text(url + "\n" + problem)
        for (url, problem) in reshape(lines, 2)
    ]
    return {p.name: p for p in problems}


def alphabetize(
    problem: JGEXFormulation,
) -> tuple[JGEXFormulation, dict[PredicateArgument, PredicateArgument]]:
    """Alphabetize the problem by renaming the points such that the points are in alphabetical order by construction order.

    Returns the renamed problem and the mapping from the alphabetized points to the original points.
    """
    original_to_alphabetized: dict[PredicateArgument, PredicateArgument] = {}

    renamed_setup_clauses = _alphabetize_clauses(
        problem.setup_clauses, original_to_alphabetized
    )
    renamed_aux_clauses = _alphabetize_clauses(
        problem.auxiliary_clauses, original_to_alphabetized
    )

    renamed_goals = tuple(
        rename_predicate_construction(goal_construction, original_to_alphabetized)
        for goal_construction in problem.goals
    )
    renamed_problem = JGEXFormulation(
        name=problem.name,
        setup_clauses=renamed_setup_clauses,
        auxiliary_clauses=renamed_aux_clauses,
        goals=renamed_goals,
    )
    alphabetized_to_original = {v: k for k, v in original_to_alphabetized.items()}
    return renamed_problem, alphabetized_to_original


def _alphabetize_clauses(
    clauses: tuple[JGEXClause, ...], mapping: dict[PredicateArgument, PredicateArgument]
) -> tuple[JGEXClause, ...]:
    for construction in clauses:
        for point in construction.points:
            if point in mapping:
                continue
            mapping[point] = PredicateArgument(ALPHABET[len(mapping)])
    return tuple(construction.renamed(mapping) for construction in clauses)



import itertools
from typing import cast

from newclid.ggb.elements_relationships import (
    BisectorLineRelationship,
    BisectorPointRelationship,
    ElementsRelationships,
    OrthoRelationship,
    ParaRelationship,
    RelationShipType,
    TangentRelationship,
)
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction


def relationships_to_newclid_assumptions(
    relationships: list[ElementsRelationships],
) -> list[PredicateConstruction]:
    assumptions: list[PredicateConstruction] = []
    for relationship in relationships:
        assumptions.extend(_relationship_to_newclid_assumptions(relationship))
    return assumptions


def _relationship_to_newclid_assumptions(
    relationship: ElementsRelationships,
) -> list[PredicateConstruction]:
    match relationship.relationship_type:
        case RelationShipType.Parallel:
            return _para_relationship_to_newclid_assumptions(relationship)
        case RelationShipType.Orthogonal:
            return _ortho_relationship_to_newclid_assumptions(relationship)
        case RelationShipType.BisectorLine:
            return _bisector_line_relationship_to_newclid_assumptions(relationship)
        case RelationShipType.BisectorPoint:
            return _bisector_point_relationship_to_newclid_assumptions(relationship)
        case RelationShipType.Tangent:
            return _tangent_relationship_to_newclid_assumptions(relationship)
    raise NotImplementedError(f"Unknown relationship: {relationship}")


def _para_relationship_to_newclid_assumptions(
    relationship: ParaRelationship,
) -> list[PredicateConstruction]:
    para_predicates: list[PredicateConstruction] = []
    for first_args in itertools.combinations(relationship.line.points, 2):
        for second_args in itertools.combinations(relationship.parallel_to.points, 2):
            first_pred_args = tuple(PredicateArgument(a) for a in first_args)
            second_pred_args = tuple(PredicateArgument(a) for a in second_args)
            para_predicates.append(
                PredicateConstruction.from_predicate_type_and_args(
                    PredicateType.PARALLEL, (*first_pred_args, *second_pred_args)
                )
            )
    return para_predicates


def _ortho_relationship_to_newclid_assumptions(
    relationship: OrthoRelationship,
) -> list[PredicateConstruction]:
    perp_predicates: list[PredicateConstruction] = []
    if len(relationship.line.points) > 1 and len(relationship.orthogonal_to.points) > 1:
        a, b = relationship.line.points[:2]
        c, d = relationship.orthogonal_to.points[:2]
        perp_predicates.append(
            PredicateConstruction.from_predicate_type_and_args(
                PredicateType.PERPENDICULAR,
                cast(tuple[PredicateArgument, ...], (a, b, c, d)),
            )
        )
    return perp_predicates


def _bisector_line_relationship_to_newclid_assumptions(
    relationship: BisectorLineRelationship,
) -> list[PredicateConstruction]:
    eqangle_predicates: list[PredicateConstruction] = []
    if (
        len(relationship.bisector_1.points) > 1
        and len(relationship.line.points) > 1
        and len(relationship.bisector_2.points) > 1
    ):
        a, b = relationship.bisector_1.points[:2]
        c, d = relationship.line.points[:2]
        g, h = relationship.bisector_2.points[:2]
        eqangle_predicates.append(
            PredicateConstruction.from_predicate_type_and_args(
                PredicateType.EQUAL_ANGLES,
                cast(tuple[PredicateArgument, ...], (a, b, c, d, c, d, g, h)),
            )
        )
    return eqangle_predicates


def _bisector_point_relationship_to_newclid_assumptions(
    relationship: BisectorPointRelationship,
) -> list[PredicateConstruction]:
    eqangle_predicates: list[PredicateConstruction] = []
    if len(relationship.line.points) > 1:
        a = relationship.angle_a.label
        b = relationship.angle_vertex.label
        c = relationship.angle_b.label
        d = next((point for point in relationship.line.points if point != b), None)
        if all(x is not None for x in (a, b, c, d)):
            eqangle_predicates.append(
                PredicateConstruction.from_predicate_type_and_args(
                    PredicateType.EQUAL_ANGLES,
                    cast(tuple[PredicateArgument, ...], (a, b, b, d, b, d, b, c)),
                )
            )
    return eqangle_predicates


def _tangent_relationship_to_newclid_assumptions(
    relationship: TangentRelationship,
) -> list[PredicateConstruction]:
    conic = relationship.tangent_to
    perp_predicates: list[PredicateConstruction] = []
    if conic.center is not None:
        for point in set(relationship.line.points).intersection(conic.points):
            point_in_line = next(
                (p for p in relationship.line.points if p != point), None
            )
            if point_in_line is not None:
                perp_predicates.append(
                    PredicateConstruction.from_predicate_type_and_args(
                        PredicateType.PERPENDICULAR,
                        cast(
                            tuple[PredicateArgument, ...],
                            (point, conic.center, point, point_in_line),
                        ),
                    )
                )
    return perp_predicates



import cProfile
import logging
from pathlib import Path

import numpy as np
import pytest
from newclid.api import GeometricSolver, GeometricSolverBuilder
from newclid.jgex.formulation import JGEXFormulation, jgex_formulation_from_txt_file
from newclid.jgex.problem_builder import JGEXProblemBuilder
from newclid.problems import ALL_PROBLEMS
from newclid.problems.problem import Problem, ProblemFormulation

from .test_jgex_builds import PROBLEMS_EXPECTED_TO_FAIL_TO_BUILD

ROOT_PATH = Path(__file__).parent.parent
DATASET_PATH = ROOT_PATH.joinpath("problems_datasets")
LOGGER = logging.getLogger(__name__)

PROBLEMS_EXPECTED_TO_FAIL = [
    "breaking_cc_tangent",
    "breaking_cc_tangent_by_kiss",
    "breaking_cc_itangent",
    "breaking_cc_itangent_by_kiss",
    # Pythagorean theorem removed for now as it produces a lot of irrelevant predicates
    # due to the irrationality result not compatible with lengths being Fractions
    # TODO: Fix pythagorean theorem
    "pyt_test_formula_to_perp",
    "pyt_test_perp_to_formula",
    # Acompute is not supported by Yuclid
    "acompute_test",
    # Cannot prescribe numerical check as goal to Yuclid
    "numerical_check_npara",
    "numerical_check_ncoll",
]

PROBLEMS_NOT_SOLVED_BEFORE = [
    "two_goals_perp_cong",
    "worlds_hardest_easy_geometry_problem1",
    "worlds_hardest_easy_geometry_problem2",
    "double_angle_implies_central_angle_2",
    "menelaus_test",
    "menelaus_frac1_test",
    "menelaus_crossed_cong_test",
    "test_l2const",
    "2020_p1",
    "2008_p6",
    "2011_p6",
    "2015_p3",
    "2021_p3",
    "1983_p2",
    "1975_p3",
]


FAILLING_PROBLEMS_NOT_EXPECTED_TO_FAIL = [
    "1995_p1",
    "2009_sl_g3_excenters",
    "imo_sl_2023_p2_constr",
    "pappus",
    "imo_sl_2023_g4_constr",
]

LARGE_EXAMPLE_PROBLEMS = jgex_formulation_from_txt_file(
    DATASET_PATH.joinpath("large_examples.txt")
)


EXAMPLE_PROBLEMS = jgex_formulation_from_txt_file(DATASET_PATH.joinpath("examples.txt"))


def _filter_problems(problem_name: str) -> None:
    if problem_name in PROBLEMS_EXPECTED_TO_FAIL_TO_BUILD:
        pytest.skip("fails to even build")
    if problem_name in PROBLEMS_EXPECTED_TO_FAIL:
        pytest.xfail("expected to fail")
    if problem_name in PROBLEMS_NOT_SOLVED_BEFORE:
        pytest.skip("not solved before")
    if problem_name in FAILLING_PROBLEMS_NOT_EXPECTED_TO_FAIL:
        pytest.skip("not expected to fail")


@pytest.mark.parametrize(
    argnames="problem",
    argvalues=EXAMPLE_PROBLEMS.values(),
    ids=EXAMPLE_PROBLEMS.keys(),
)
def test_example_problems(problem: JGEXFormulation):
    _filter_problems(problem.name)
    match_logger = logging.getLogger("newclid.match_theorems")
    match_logger.setLevel(logging.INFO)
    rng = np.random.default_rng(42)
    problem_builder = JGEXProblemBuilder(rng).with_problem(problem)
    solver = GeometricSolverBuilder(rng).build(problem_builder.build())
    success = solver.run()
    assert success


@pytest.mark.slow
@pytest.mark.parametrize(
    argnames="problem",
    argvalues=LARGE_EXAMPLE_PROBLEMS.values(),
    ids=LARGE_EXAMPLE_PROBLEMS.keys(),
)
def test_large_example_problems(problem: JGEXFormulation):
    _filter_problems(problem.name)
    rng = np.random.default_rng(42)
    problem_builder = JGEXProblemBuilder(rng).with_problem(problem)
    solver: GeometricSolver = GeometricSolverBuilder(rng).build(problem_builder.build())
    match_logger = logging.getLogger("newclid.match_theorems")
    match_logger.setLevel(logging.INFO)
    profile = False
    if profile:
        prof = cProfile.Profile()
        success = prof.runcall(solver.run)
        profilings_path = Path(__file__).parent.joinpath(
            "large_example_problems_profiling"
        )
        profilings_path.mkdir(parents=True, exist_ok=True)
        profilings_file = str(profilings_path.joinpath(f"{problem.name}.prof"))
        prof.dump_stats(profilings_file)
    else:
        success = solver.run()

    print(solver.run_infos)
    assert success


IMO_PROBLEMS = jgex_formulation_from_txt_file(DATASET_PATH.joinpath("imo.txt"))


@pytest.mark.slow
@pytest.mark.parametrize(
    argnames="problem",
    argvalues=IMO_PROBLEMS.values(),
    ids=IMO_PROBLEMS.keys(),
)
def test_imo(problem: JGEXFormulation):
    _filter_problems(problem.name)
    rng = np.random.default_rng(42)
    jgex_problem_setup = (
        JGEXProblemBuilder(rng)
        .with_problem(problem)
        .include_auxiliary_clauses()
        .build()
    )
    solver: GeometricSolver = GeometricSolverBuilder(rng).build(jgex_problem_setup)
    match_logger = logging.getLogger("newclid.match_theorems")
    match_logger.setLevel(logging.INFO)
    success = solver.run()
    assert success


IMO_PROBLEMS_SL = jgex_formulation_from_txt_file(DATASET_PATH.joinpath("imo_sl.txt"))


@pytest.mark.slow
@pytest.mark.parametrize(
    argnames="problem",
    argvalues=IMO_PROBLEMS_SL.values(),
    ids=IMO_PROBLEMS_SL.keys(),
)
def test_imo_sl(problem: JGEXFormulation):
    _filter_problems(problem.name)
    rng = np.random.default_rng(42)
    jgex_problem_setup = (
        JGEXProblemBuilder(rng)
        .with_problem(problem)
        .include_auxiliary_clauses()
        .build()
    )
    solver: GeometricSolver = GeometricSolverBuilder(rng).build(jgex_problem_setup)
    match_logger = logging.getLogger("newclid.match_theorems")
    match_logger.setLevel(logging.INFO)
    success = solver.run()
    assert success


@pytest.mark.parametrize(
    argnames="problem,formulation",
    argvalues=[
        (problem, formulation)
        for problem in ALL_PROBLEMS
        for formulation in problem.formulations
    ],
    ids=[
        f"{problem.name} | {formulation.name}"
        for problem in ALL_PROBLEMS
        for formulation in problem.formulations
    ],
)
def test_problem_constants(problem: Problem, formulation: ProblemFormulation):
    jgex_logger = logging.getLogger("newclid.jgex")
    jgex_logger.setLevel(logging.INFO)
    match_logger = logging.getLogger("newclid.match_theorems")
    match_logger.setLevel(logging.INFO)

    rng = np.random.default_rng(42)

    formulation_version = formulation.formulation
    match formulation_version.formulation_type:
        case "jgex":
            problem_setup = (
                JGEXProblemBuilder(rng, max_attempts_per_clause=10)
                .with_problem(formulation_version)
                .include_auxiliary_clauses()
                .build(max_attempts_to_satisfy_goals_numerically=100)
            )
    solver: GeometricSolver = GeometricSolverBuilder(rng).build(problem_setup)

    success = solver.run()
    assert success



"""Definitions that will add points in order to complete a geometric figure"""

from newclid.jgex.clause import JGEXClause
from newclid.jgex.constructions._index import JGEXConstructionName
from newclid.jgex.definition import JGEXDefinition, SketchConstruction, v
from newclid.jgex.sketch import (
    sketch_bline,
    sketch_centroid,
    sketch_circle,
    sketch_function_name,
    sketch_isosvertex,
    sketch_pline,
    sketch_rotaten90,
    sketch_rotatep90,
    sketch_square,
)

ISO_TRIANGLE_VERTEX = JGEXDefinition(
    name=JGEXConstructionName.ISO_TRIANGLE_VERTEX,
    args=(v("x"), v("b"), v("c")),
    rely_on_points={v("x"): (v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c")[0],
    clauses=JGEXClause.from_str("cong x b x c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_isosvertex.__name__),
            args=(v("b"), v("c")),
        ),
    ),
    input_points=(v("b"), v("c")),
    output_points=(v("x"),),
)
ISO_TRIANGLE_VERTEX_ANGLE = JGEXDefinition(
    name=JGEXConstructionName.ISO_TRIANGLE_VERTEX_ANGLE,
    args=(v("x"), v("b"), v("c")),
    rely_on_points={v("x"): (v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c")[0],
    clauses=JGEXClause.from_str("eqangle x b b c b c x c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_isosvertex.__name__),
            args=(v("b"), v("c")),
        ),
    ),
    input_points=(v("b"), v("c")),
    output_points=(v("x"),),
)

EQ_TRIANGLE = JGEXDefinition(
    name=JGEXConstructionName.EQ_TRIANGLE,
    args=(v("x"), v("b"), v("c")),
    rely_on_points={v("x"): (v("b"), v("c"))},
    requirements=JGEXClause.from_str("diff b c")[0],
    clauses=JGEXClause.from_str(
        "cong x b b c, cong b c c x; eqangle b x b c c b c x, eqangle x c x b b x b c"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("b"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_circle.__name__),
            args=(v("c"), v("b"), v("c")),
        ),
    ),
    input_points=(v("b"), v("c")),
    output_points=(v("x"),),
)

ISO_TRAPEZOID2 = JGEXDefinition(
    name=JGEXConstructionName.ISO_TRAPEZOID2,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("cong x b a c, cong x a b c"),
    sketches=(
        SketchConstruction(name="iso_trapezoid2", args=(v("a"), v("b"), v("c"))),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

PARALLELOGRAM = JGEXDefinition(
    name=JGEXConstructionName.PARALLELOGRAM,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "para a b c x, para a x b c; cong a b c x, cong a x b c"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_pline.__name__),
            args=(v("c"), v("a"), v("b")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

SQUARE = JGEXDefinition(
    name=JGEXConstructionName.SQUARE,
    args=(v("x"), v("y"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b")), v("y"): (v("a"), v("b"), v("x"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str(
        "x : perp a b b x, cong a b b x; y : para a b x y, para a y b x; perp a y y x, cong b x x y, cong x y y a, perp a x b y, cong a x b y"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_square.__name__), args=(v("a"), v("b"))
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"), v("y")),
)


CIRCLE = JGEXDefinition(
    name=JGEXConstructionName.CIRCLE,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("circle x a b c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_bline.__name__), args=(v("a"), v("b"))
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_bline.__name__), args=(v("a"), v("c"))
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

CIRCUMCENTER = JGEXDefinition(
    name=JGEXConstructionName.CIRCUMCENTER,
    args=(v("x"), v("a"), v("b"), v("c")),
    rely_on_points={v("x"): (v("a"), v("b"), v("c"))},
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str("cong x a x b, cong x b x c"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_bline.__name__), args=(v("a"), v("b"))
        ),
        SketchConstruction(
            name=sketch_function_name(sketch_bline.__name__), args=(v("a"), v("c"))
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)

PSQUARE = JGEXDefinition(
    name=JGEXConstructionName.PSQUARE,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("cong x a a b, perp x a a b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_rotatep90.__name__), args=(v("a"), v("b"))
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)


NSQUARE = JGEXDefinition(
    name=JGEXConstructionName.NSQUARE,
    args=(v("x"), v("a"), v("b")),
    rely_on_points={v("x"): (v("a"), v("b"))},
    requirements=JGEXClause.from_str("diff a b")[0],
    clauses=JGEXClause.from_str("cong x a a b, perp x a a b"),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_rotaten90.__name__),
            args=(v("a"), v("b")),
        ),
    ),
    input_points=(v("a"), v("b")),
    output_points=(v("x"),),
)

CENTROID = JGEXDefinition(
    name=JGEXConstructionName.CENTROID,
    args=(v("x"), v("y"), v("z"), v("i"), v("a"), v("b"), v("c")),
    rely_on_points={
        v("x"): (v("b"), v("c")),
        v("y"): (v("c"), v("a")),
        v("z"): (v("a"), v("b")),
        v("i"): (v("a"), v("x"), v("b"), v("y")),
    },
    requirements=JGEXClause.from_str("ncoll a b c")[0],
    clauses=JGEXClause.from_str(
        "x : coll x b c, cong x b x c; y : coll y c a, cong y c y a; z : coll z a b, cong z a z b; i : coll a x i, coll b y i; coll c z i"
    ),
    sketches=(
        SketchConstruction(
            name=sketch_function_name(sketch_centroid.__name__),
            args=(v("a"), v("b"), v("c")),
        ),
    ),
    input_points=(v("a"), v("b"), v("c")),
    output_points=(v("x"),),
)


COMPLETE_FORM_CONSTRUCTIONS = [
    ISO_TRIANGLE_VERTEX,
    ISO_TRIANGLE_VERTEX_ANGLE,
    EQ_TRIANGLE,
    ISO_TRAPEZOID2,
    PARALLELOGRAM,
    SQUARE,
    CIRCLE,
    CIRCUMCENTER,
    PSQUARE,
    NSQUARE,
    CENTROID,
]



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from newclid.numerical import close_enough
from newclid.numerical.geometries import circle_num_from_points_around
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Point

if TYPE_CHECKING:
    from newclid.proof_state import ProofState


class Cyclic(PredicateInterface):
    """cyclic A B C D -
    Represent that 4 (or more) points lie on the same circle."""

    predicate_type: Literal[PredicateType.CYCLIC] = PredicateType.CYCLIC

    points: tuple[Point, ...]

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        if len(set(args)) <= 3:
            return None
        return tuple(sorted(set(args)))

    def check_numerical(self) -> bool:
        p1, p2, p3 = self.points[:3]
        try:
            circle = circle_num_from_points_around([p1.num, p2.num, p3.num])
        except ValueError:
            return False
        return all(
            close_enough(circle.radius**2, circle.center.distance2(p.num))
            for p in self.points[3:]
        )

    def check(self, proof_state: ProofState) -> bool | None:
        return proof_state.symbols.circles.check_cyclic(self.points)

    def add(self, proof_state: ProofState) -> tuple[PredicateInterface, ...]:
        proof_state.symbols.circles.make_cyclic(self.points, justification=self)
        return ()

    def to_constructive(self, point: str) -> str:
        a, b, c = [x.name for x in self.points if x.name != point]
        return f"on_circum {point} {a} {b} {c}"

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.points)

    def __str__(self) -> str:
        return f"{''.join(str(p) for p in self.points)} are cyclic"



import logging
from pathlib import Path
from typing import Self
from xml.etree.ElementTree import parse
from zipfile import ZipFile

from pydantic import BaseModel

from newclid.api import ProblemBuilder
from newclid.ggb.apply_commands_effects import apply_commands_effects_on_elements
from newclid.ggb.assumptions_from_commands import commands_to_newclid_assumptions
from newclid.ggb.assumptions_from_elements import elements_to_newclid_assumptions
from newclid.ggb.assumptions_from_elements_relation import (
    relationships_to_newclid_assumptions,
)
from newclid.ggb.elements_relationships import ElementsRelationships
from newclid.ggb.read_commands import GGBCommand, read_commands
from newclid.ggb.read_elements import GGBConic, GGBLine, GGBPoint, read_elements
from newclid.ggb.relationships_from_commands import relationships_from_commands
from newclid.problem import PredicateConstruction, ProblemSetup
from newclid.symbols.points_registry import Point

LOGGER = logging.getLogger(__name__)


class GeogebraProblemBuilder(ProblemBuilder):
    def __init__(self, ggb_file_path: Path):
        self.ggb_file_path = ggb_file_path
        self.problem_name: str = "GeogebraProblem"
        self.goals: list[PredicateConstruction] = []
        self.removed_assumptions: list[PredicateConstruction] = []
        self.extra_assumptions: list[PredicateConstruction] = []

    def with_goals(self, goals: list[PredicateConstruction]) -> Self:
        self.goals.extend(goals)
        return self

    def without_assumptions(self, assumptions: list[PredicateConstruction]) -> Self:
        self.removed_assumptions.extend(assumptions)
        return self

    def with_extra_assumptions(self, assumptions: list[PredicateConstruction]) -> Self:
        self.extra_assumptions.extend(assumptions)
        return self

    def build(self) -> ProblemSetup:
        problem_name = self.ggb_file_path.name.removesuffix(".ggb")
        points, assumptions = load_ggb_file(self.ggb_file_path)
        assumptions.extend(self.extra_assumptions)
        assumptions = [a for a in assumptions if a not in self.removed_assumptions]
        problem = ProblemSetup(
            name=problem_name,
            goals=tuple(self.goals),
            points=tuple(points),
            assumptions=tuple(assumptions),
        )
        LOGGER.info(
            f"Loaded problem ’{problem.name}’ with {len(points)} points"
            f" and {len(assumptions)} assumptions from GeoGebra."
        )
        return problem


def load_ggb_file(
    ggb_file_path: Path,
) -> tuple[list[Point], list[PredicateConstruction]]:
    ggbsetup = _load_geogebra_setup(ggb_file_path)

    points = [p.to_newclid() for p in ggbsetup.points.values()]

    assumptions: list[PredicateConstruction] = []
    assumptions.extend(
        commands_to_newclid_assumptions(
            ggbsetup.commands, ggbsetup.lines, ggbsetup.conics
        )
    )
    assumptions.extend(elements_to_newclid_assumptions(ggbsetup.lines, ggbsetup.conics))
    assumptions.extend(relationships_to_newclid_assumptions(ggbsetup.relationships))
    points.sort(key=lambda p: p.name)
    deduping_assumptions = set(assumptions)
    assumptions = list(deduping_assumptions)
    return points, assumptions


class GGBSetup(BaseModel):
    points: dict[str, GGBPoint]
    lines: dict[str, GGBLine]
    conics: dict[str, GGBConic]
    commands: list[GGBCommand]
    relationships: list[ElementsRelationships]


def _load_geogebra_setup(ggb_file_path: Path) -> GGBSetup:
    LOGGER.debug(f"Loading GeoGebra file at {ggb_file_path}")
    with ZipFile(ggb_file_path, "r").open("geogebra.xml") as ggb:
        tree = parse(ggb)
    root = tree.getroot()
    points, lines, conics = read_elements(root)
    points_str = "\n" + "\n".join(
        pt.model_dump_json(indent=2) for pt in points.values()
    )
    lines_str = "\n" + "\n".join(
        line.model_dump_json(indent=2) for line in lines.values()
    )
    conics_str = "\n" + "\n".join(
        conic.model_dump_json(indent=2) for conic in conics.values()
    )
    LOGGER.debug(
        f"GGB Elements:\nPoints: {points_str}\nLines: {lines_str}\nConics: {conics_str}"
    )

    commands = read_commands(root, points, lines, conics)
    commands_str = "\n" + "\n".join(c.model_dump_json(indent=2) for c in commands)
    LOGGER.debug(f"GGB Commands: {commands_str}")
    apply_commands_effects_on_elements(commands, lines, conics)
    relationships = relationships_from_commands(commands, points, lines, conics)
    return GGBSetup(
        points=points,
        lines=lines,
        conics=conics,
        commands=commands,
        relationships=relationships,
    )



from __future__ import annotations

from typing import Collection, Iterator, TypeAlias

from pydantic import BaseModel

from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument


class Point(BaseModel):
    """The Point class is the symbolic representation of a geometric point."""

    name: PredicateArgument
    num: PointNum
    """The numerical representation of the point."""

    def __str__(self) -> str:
        p = self.name[0].upper()
        for c in self.name[1:]:
            if c.isdigit():
                p += chr(ord("₀") + ord(c) - ord("0"))
            else:
                p += f"_{c}"
        return p

    def __hash__(self) -> int:
        return hash(str(self))


class PointsRegisty:
    """Registry of points, used in the proof state to manage existing points."""

    def __init__(self) -> None:
        self.name_to_point: dict[PredicateArgument, Point] = {}

    def add_point(self, point: Point) -> None:
        self.name_to_point[point.name] = point

    def names2points(self, pnames: Collection[PredicateArgument]) -> list[Point]:
        """Return Point objects given names."""
        points: list[Point] = []
        for name in pnames:
            if name not in self.name_to_point:
                raise ValueError(f"Cannot find point {name} in graph")
            points.append(self.name_to_point[name])
        return points

    def __iter__(self) -> Iterator[Point]:
        return iter(self.name_to_point.values())


"""
Here we list the different types corresponding to geometric objects that are arguments of predicates. They are all composed of instances of the Point class. They are:

- Segment: a 2-tuple of Points representing a segment with the points are extremities.
- Ratio: a 2-tuple of Segments representing a ratio between the lengths of two segments.
- Line: a 2-tuple of Points representing a line through them.
- Angle: a 2-tuple of Lines representing the angle formed by them.
- Triangle: a 3-tuple of Points representing a triangle with the points as vertices.
"""

Segment: TypeAlias = tuple[Point, Point]
Ratio: TypeAlias = tuple[Segment, Segment]
Line: TypeAlias = tuple[Point, Point]
Angle: TypeAlias = tuple[Line, Line]
Triangle: TypeAlias = tuple[Point, Point, Point]



from __future__ import annotations

from fractions import Fraction
from typing import Literal

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Segment
from newclid.tools import fraction_to_ratio, str_to_fraction


class SquaredConstantRatio(PredicateInterface):
    """r2const A B C D r -
    Represent that |AB|² / |CD|² = r

    r should be given with numerator and denominator separated by '/', as in 2/3.
    """

    predicate_type: Literal[PredicateType.SQUARED_CONSTANT_RATIO] = (
        PredicateType.SQUARED_CONSTANT_RATIO
    )
    segment1: Segment
    segment2: Segment
    square_ratio: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        try:
            a, b, c, d, r = args
        except ValueError:
            raise ValueError(f"Invalid arguments: {args}")
        if a == b or c == d:
            return None
        f = str_to_fraction(r)
        a, b = sorted((a, b))
        c, d = sorted((c, d))
        if (a, b) > (c, d):
            a, b, c, d = c, d, a, b
            f = 1 / f
        return (a, b, c, d, PredicateArgument(fraction_to_ratio(f)))

    def check_numerical(self) -> bool:
        a, b = self.segment1
        c, d = self.segment2
        r = self.square_ratio
        numerator = (a.num - b.num).dot(a.num - b.num)
        denominator = (c.num - d.num).dot(c.num - d.num)
        return close_enough(numerator / denominator, float(r))

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        a, b = self.segment1
        c, d = self.segment2
        r = self.square_ratio
        return (
            PredicateArgument(a.name),
            PredicateArgument(b.name),
            PredicateArgument(c.name),
            PredicateArgument(d.name),
            PredicateArgument(fraction_to_ratio(r)),
        )

    def __str__(self) -> str:
        a, b = self.segment1
        c, d = self.segment2
        r = self.square_ratio
        return f"|{a}{b}|\u00b2:|{c}{d}|\u00b2 = {fraction_to_ratio(r)}"



"""Classical Breadth-First Search based agents."""

from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any, Callable, NamedTuple, Optional

from newclid.agent.agents_interface import DeductiveAgent
from newclid.agent.ddarn import DDARN
from newclid.justifications.justification import justify_dependency
from newclid.predicate_types import PredicateArgument
from newclid.predicates import predicate_from_construction
from newclid.problem import PredicateConstruction
from newclid.proof_state import ProofState
from newclid.rule import Rule
from newclid.tools import atomize


class NamedFunction(NamedTuple):
    name: str
    function: Callable[..., Any]


class HumanAgent(DeductiveAgent):
    def __init__(self) -> None:
        self.ddarn: Optional[DDARN] = None
        self.server: Optional[Path] = None

    @classmethod
    def select(
        cls,
        options: list[NamedFunction],
        allow_all: bool = False,
        allow_none: bool = False,
    ) -> NamedFunction | None:
        options_string = "\nSelect an option:"
        for i, option in enumerate(options):
            options_string += f"\n- [{i}] {option.name}"
        if allow_all:
            options_string += "- [all]"
        if allow_none:
            options_string += "- [none]"

        options_string += "\nChoose:"
        choice = input(options_string)
        if choice == "all":
            for option in options:
                option.function()
            return None
        if choice == "none":
            return None
        n = int(choice)
        return options[n]

    def step(self, proof: ProofState, rules: list[Rule]) -> bool:
        exausted = False
        match_rules = [
            NamedFunction(
                f"match {theorem.fullname}: {theorem}",
                self._fn_match(proof, theorem),
            )
            for theorem in rules
        ]
        if self.ddarn:
            if not self.ddarn.step(proof=proof, rules=rules):
                self.ddarn = None
        else:
            print("Premises:")
            for dep in proof.graph.premises():
                print(f"{dep.predicate}")
            selected = self.select(
                [
                    NamedFunction(
                        "match", partial(self.match, match_rules=match_rules)
                    ),
                    NamedFunction(
                        "update server", partial(self.update_server, proof=proof)
                    ),
                    NamedFunction(
                        "construction", partial(self.add_construction, proof=proof)
                    ),
                    NamedFunction("exhaust with ddarn", self.exhaust_with_ddarn),
                    NamedFunction("check predicate", partial(self.check, proof=proof)),
                    NamedFunction(
                        "check goals", partial(self.check_goals, proof=proof)
                    ),
                    NamedFunction("stop", lambda: None),
                ],
                allow_none=True,
            )
            if selected is None:
                return not exausted
            selected.function()
            if selected.name == "stop":
                exausted = True
        if proof.check_goals():
            print("All the goals are proven")
        else:
            for goal in proof.goals:
                print(f"{goal} proven? {proof.check(goal)}")
        return not exausted

    def match(self, match_rules: list[NamedFunction]) -> None:
        selected = self.select(match_rules, allow_none=True)
        if selected is not None:
            selected.function()

    def update_server(self, proof: "ProofState") -> None:
        self._pull_to_server(proof)

    def add_construction(self, proof: "ProofState") -> None:
        raise NotImplementedError

    def exhaust_with_ddarn(self) -> None:
        self.ddarn = DDARN()

    def check(self, proof: "ProofState") -> None:
        user_input: tuple[str, *tuple[PredicateArgument, ...]] = atomize(
            input("Check: ")
        )  # type: ignore
        predicate_construction = PredicateConstruction.from_tuple(user_input)
        predicate = predicate_from_construction(
            predicate_construction, proof.symbols.points
        )
        if predicate is None:
            print("predicate not parsed")
        else:
            print(f"Check result of {predicate}: {proof.check(predicate)}")

    def check_goals(self, proof: "ProofState") -> None:
        selects: list[NamedFunction] = [
            NamedFunction(
                f"check {goal}",
                lambda: print(f"Check result of {goal}: {proof.check(goal)}"),
            )
            for goal in proof.goals
        ]
        selected = self.select(selects, allow_all=True, allow_none=True)
        if selected is not None:
            selected.function()

    def _pull_to_server(self, proof: "ProofState"):  # type: ignore
        raise NotImplementedError
        # assert proof.problem_path
        # if not self.server:
        #     self.server = proof.problem_path / "html"
        #     run_static_server(self.server)
        # webapp.pull_to_server(proof, server_path=self.server)

    def _fn_match(self, proof: "ProofState", rule: Rule):
        def fn():
            deps = list(proof.match_theorem(rule))
            selected = self.select(
                [
                    NamedFunction(
                        f"{', '.join(str(s) for s in justify_dependency(dep, proof))} => {dep.predicate}",
                        lambda dep=dep: proof.apply(dep),
                    )
                    for dep in deps
                ],
                allow_all=True,
                allow_none=True,
            )
            if selected is None:
                return
            return selected.function

        return fn



from newclid.jgex.formulation import ALPHABET
from newclid.predicate_types import PredicateArgument


def get_available_from_alphabet(
    current_points: list[PredicateArgument],
) -> PredicateArgument | None:
    """
    Get the first available symbol in the alphabet not used in list.

    Args:
        list (List[str]): A list of symbols to check (typically point names).

    Returns:
        str: The first available symbol from the list.
    """
    for symbol in ALPHABET:
        if symbol not in current_points:
            return PredicateArgument(symbol)
    return None



import numpy as np
import pytest
from newclid.api import GeometricSolverBuilder, PythonDefault
from newclid.deductors.sympy_ar.algebraic_manipulator import SympyARDeductor
from newclid.jgex.problem_builder import JGEXProblemBuilder


class TestDDAR:
    @pytest.fixture(autouse=True)
    def setUpClass(self):
        self.rng = np.random.default_rng(998244353)
        self.solver_builder = GeometricSolverBuilder(
            rng=self.rng, api_default=PythonDefault(use_sympy_ar=True)
        )
        self.problem_builder = JGEXProblemBuilder(rng=self.rng)

    def test_incenter_excenter_should_succeed(self):
        problem_setup = self.problem_builder.with_problem_from_txt(
            "a b c = triangle a b c; d = incenter d a b c; e = excenter e a b c ? perp d c c e"
        ).build()
        solver = (
            self.solver_builder.with_deductors([SympyARDeductor()])
            .with_rules([])
            .build(problem_setup)
        )
        success = solver.run()
        assert success

    def test_orthocenter_should_succeed(self):
        problem_setup = self.problem_builder.with_problem_from_txt(
            "a b c = triangle a b c; d = on_tline d b a c, on_tline d c a b ? perp a d b c"
        ).build()
        solver = self.solver_builder.with_deductors([SympyARDeductor()]).build(
            problem_setup
        )
        success = solver.run()
        assert success
        # solver.write_proof_steps(Path(r"./tests_output/orthocenter_proof.txt"))



from typing import Annotated

from pydantic import Field

from newclid.heuristics._index import HeuristicName
from newclid.heuristics._interface import Heuristic
from newclid.heuristics.angle_vertices import (
    AngleVerticesHeuristic,
    AngleVerticesHeuristicConfig,
)
from newclid.heuristics.centers_of_cyclic import (
    CentersOfCyclicHeuristic,
    CentersOfCyclicHeuristicConfig,
)
from newclid.heuristics.circumcircles_for_eqangle import (
    ThreeCircumcirclesForEqangleGoalHeuristic,
    ThreeCircumcirclesForEqangleGoalHeuristicConfig,
)
from newclid.heuristics.foots_on_lines import FootHeuristic, FootHeuristicConfig
from newclid.heuristics.line_intersections import (
    LineIntersectionsHeuristic,
    LineIntersectionsHeuristicConfig,
)
from newclid.heuristics.midpoint import MidpointHeuristic, MidpointHeuristicConfig
from newclid.heuristics.reflect_on_center import (
    ReflectOnCenterHeuristic,
    ReflectOnCenterHeuristicConfig,
)
from newclid.heuristics.transfer_distances import (
    TransferDistancesHeuristic,
    TransferDistancesHeuristicConfig,
)

HeuristicConfig = Annotated[
    MidpointHeuristicConfig
    | LineIntersectionsHeuristicConfig
    | TransferDistancesHeuristicConfig
    | AngleVerticesHeuristicConfig
    | ReflectOnCenterHeuristicConfig
    | CentersOfCyclicHeuristicConfig
    | ThreeCircumcirclesForEqangleGoalHeuristicConfig
    | FootHeuristicConfig,
    Field(discriminator="heuristic_name"),
]


def heuristic_from_name(name: HeuristicName) -> Heuristic:
    return heuristic_from_config(config_from_name(name))


def config_from_name(name: HeuristicName) -> HeuristicConfig:
    match name:
        case HeuristicName.MIDPOINT:
            return MidpointHeuristicConfig()
        case HeuristicName.LINE_INTERSECTIONS:
            return LineIntersectionsHeuristicConfig()
        case HeuristicName.EQDISTANCE:
            return TransferDistancesHeuristicConfig()
        case HeuristicName.ANGLE_VERTICES:
            return AngleVerticesHeuristicConfig()
        case HeuristicName.THREE_CIRCUMCIRCLES_FOR_EQANGLE_GOAL:
            return ThreeCircumcirclesForEqangleGoalHeuristicConfig()
        case HeuristicName.REFLECT_ON_CENTER:
            return ReflectOnCenterHeuristicConfig()
        case HeuristicName.CENTERS:
            return CentersOfCyclicHeuristicConfig()
        case HeuristicName.FOOT:
            return FootHeuristicConfig()
    raise ValueError(f"Unknown heuristic: {name}")


def heuristic_from_config(
    heuritic_config: HeuristicConfig,
) -> Heuristic:
    match heuritic_config.heuristic_name:
        case HeuristicName.MIDPOINT:
            return MidpointHeuristic()
        case HeuristicName.LINE_INTERSECTIONS:
            return LineIntersectionsHeuristic()
        case HeuristicName.EQDISTANCE:
            return TransferDistancesHeuristic()
        case HeuristicName.ANGLE_VERTICES:
            return AngleVerticesHeuristic()
        case HeuristicName.THREE_CIRCUMCIRCLES_FOR_EQANGLE_GOAL:
            return ThreeCircumcirclesForEqangleGoalHeuristic()
        case HeuristicName.REFLECT_ON_CENTER:
            return ReflectOnCenterHeuristic()
        case HeuristicName.CENTERS:
            return CentersOfCyclicHeuristic()
        case HeuristicName.FOOT:
            return FootHeuristic()
    raise ValueError(f"Unknown heuristic: {heuritic_config.heuristic_name}")



from typing import TypedDict

from matplotlib import patches
from matplotlib.artist import Artist
from matplotlib.axes import Axes

from newclid.numerical.geometries import PointNum
from newclid.symbols.circles_registry import CircleSymbol


class ArtistKwargs(TypedDict, total=False):
    color: str
    lw: float
    linewidth: float
    linestyle: str
    alpha: float
    zorder: int
    arrowstyle: str | patches.ArrowStyle
    width: float
    height: float
    fill: bool


def draw_circle_symbol(
    ax: Axes,
    circle: CircleSymbol,
    line_color: str,
    line_width: float,
) -> Artist:
    return draw_circle(
        ax,
        (circle.num.center.x, circle.num.center.y),
        circle.num.radius,
        line_color,
        line_width,
    )


def draw_circle(
    ax: Axes,
    center: tuple[float, float],
    radius: float,
    line_color: str,
    line_width: float,
) -> Artist:
    circle = patches.Circle(
        center,
        radius,
        color=line_color,
        fill=False,
        lw=line_width,
    )
    ax.add_artist(circle)
    return circle


def draw_triangle(
    ax: Axes,
    p0: PointNum,
    p1: PointNum,
    p2: PointNum,
    line_color: str,
    line_width: float,
) -> Artist:
    triangle = patches.Polygon(
        ((p0.x, p0.y), (p1.x, p1.y), (p2.x, p2.y)),
        closed=True,
        zorder=15,
        color=line_color,
        lw=line_width,
        fill=False,
    )
    ax.add_artist(triangle)
    return triangle


def draw_segment(
    ax: Axes,
    p0: PointNum,
    p1: PointNum,
    line_color: str,
    line_width: float,
) -> Artist:
    segment = patches.Polygon(
        ((p0.x, p0.y), (p1.x, p1.y)),
        closed=False,
        zorder=15,
        color=line_color,
        lw=line_width,
        fill=False,
    )
    ax.add_artist(segment)
    return segment


def draw_arrow(
    ax: Axes,
    p0: PointNum,
    p1: PointNum,
    line_color: str,
    line_width: float,
) -> Artist:
    midpoint = (p0 + p1) / 2
    arrow = patches.FancyArrowPatch(
        (p0.x, p0.y),
        (midpoint.x, midpoint.y),
        zorder=5,
        color=line_color,
        lw=line_width,
        arrowstyle="simple,head_length=10,head_width=10",
    )
    ax.add_artist(arrow)
    return arrow



from typing import Any

from newclid.agent._index import AgentName
from newclid.agent.agents_interface import DeductiveAgent
from newclid.agent.ddarn import DDARN
from newclid.agent.follow_deductions import FollowDeductions
from newclid.agent.human_agent import HumanAgent


def make_agent(agent_name: str | AgentName, **kwargs: Any) -> DeductiveAgent:
    match AgentName(agent_name):
        case AgentName.DDARN:
            return DDARN()
        case AgentName.HUMAN_AGENT:
            return HumanAgent()
        case AgentName.FOLLOW_DEDUCTIONS:
            return FollowDeductions(deductions_provider=kwargs["deductions_provider"])



"""Helper functions to write proofs in a natural language."""

from __future__ import annotations

from collections import defaultdict
from fractions import Fraction

from pydantic import BaseModel

from newclid.deductors.deductor_interface import ARDeduction, ARPremise
from newclid.justifications._index import JustificationType
from newclid.justifications.justification import Justification
from newclid.predicates import Predicate
from newclid.proof_data import PredicateInProof, ProofData
from newclid.rule import RuleApplication


class ProofSections(BaseModel):
    points: list[str]
    assumptions: list[str]
    numerical_checks: list[str]
    trivial_predicates: list[str]
    proven_goals: list[str]
    unproven_goals: list[str]
    proof_steps: list[str]
    appendix_ar: list[str]


def write_proof(proof_data: ProofData) -> str:
    """Output the solution to out_file.

    Args:
      proof_state: Proof state.
      problem: Containing the problem definition and theorems.
      out_file: file to write to, empty string to skip writing to file.
    """
    sections = write_proof_sections(proof_data)
    proof_lines: list[str] = ["# Problem setup:"]
    proof_lines.append("\n## Points")
    proof_lines.extend(sections.points)
    proof_lines.append("\n## Assumptions:\n")
    proof_lines.extend(sections.assumptions)
    if sections.numerical_checks:
        proof_lines.append("\n## Numerical checks\n")
        proof_lines.extend(sections.numerical_checks)

    proof_lines.append("\n\n# Goals\n")
    proof_lines.extend(sections.proven_goals)
    if sections.unproven_goals:
        proof_lines.append("\n## Unproven goals")
        proof_lines.extend(sections.unproven_goals)

    proof_lines.append("\n\n# Proof:\n")
    proof_lines.extend(sections.proof_steps)

    proof_lines.append("\n\n# Appendix: AR deductions:")
    proof_lines.extend(sections.appendix_ar)

    proof_lines.append("\nEnd of proof.")

    return "\n".join(proof_lines)


def write_proof_sections(proof_data: ProofData) -> ProofSections:
    point_prefix = "\n- "
    point_lines: list[str] = [
        f"{point_prefix}{point_prefix.join(f'{p}({p.num.x:.4f}, {p.num.y:.4f})' for p in proof_data.points)}"
    ]
    assumption_lines: list[str] = [
        f"[{assumption.id}] : {assumption.predicate}"
        for assumption in proof_data.construction_assumptions
    ]
    numerical_check_lines: list[str] = [
        f"[{numerical_check.id}] : {numerical_check.predicate}"
        for numerical_check in proof_data.numerical_checks
    ]
    trivial_predicates_lines: list[str] = [
        f"[{trivial_predicate.id}] : {trivial_predicate.predicate}"
        for trivial_predicate in proof_data.trivial_predicates
    ]

    proven_goal_lines: list[str] = [
        f"{proven_goal.predicate} : Proved [{proven_goal.id}]"
        for proven_goal in proof_data.proven_goals
    ]
    unproven_goal_lines: list[str] = [
        f"{unproven_goal} : Could not prove"
        for unproven_goal in proof_data.unproven_goals
    ]

    predicate_to_id: dict[Predicate, PredicateInProof] = {}
    for predicate_in_proof in (
        proof_data.construction_assumptions
        + proof_data.numerical_checks
        + proof_data.trivial_predicates
    ):
        predicate_to_id[predicate_in_proof.predicate] = predicate_in_proof

    for proof_step in proof_data.proof_steps:
        predicate_to_id[proof_step.proven_predicate.predicate] = (
            proof_step.proven_predicate
        )

    ar_deductions: list[tuple[int, ARDeduction]] = []
    proof_lines: list[str] = []
    for step_id, proof_step in enumerate(proof_data.proof_steps):
        line = _write_justification_line(
            justification=proof_step.justification,
            predicate_to_id=predicate_to_id,
            ar_deductions=ar_deductions,
            proven_predicate=proof_step.proven_predicate,
            step_id=step_id,
        )
        proof_lines.append(f"{step_id:03d}. | {line}")

    appendix_ar_lines: list[str] = []
    for step_id, ar_deduction in ar_deductions:
        appendix_ar_lines.append("\n\n")
        appendix_ar_lines.extend(_write_ar_deduction_appendix(step_id, ar_deduction))

    return ProofSections(
        points=point_lines,
        assumptions=assumption_lines,
        numerical_checks=numerical_check_lines,
        trivial_predicates=trivial_predicates_lines,
        proven_goals=proven_goal_lines,
        unproven_goals=unproven_goal_lines,
        proof_steps=proof_lines,
        appendix_ar=appendix_ar_lines,
    )


def _write_justification_line(
    justification: Justification,
    predicate_to_id: dict[Predicate, PredicateInProof],
    ar_deductions: list[tuple[int, ARDeduction]],
    proven_predicate: PredicateInProof,
    step_id: int,
) -> str | None:
    match justification.dependency_type:
        case (
            JustificationType.ASSUMPTION
            | JustificationType.NUMERICAL_CHECK
            | JustificationType.REFLEXIVITY
        ):
            return None
        case JustificationType.RULE_APPLICATION:
            return _rule_application_to_line(
                proven_predicate=proven_predicate,
                rule_application=justification,
                predicate_to_id=predicate_to_id,
            )
        case JustificationType.AR_DEDUCTION:
            ar_deductions.append((step_id, justification))
            return _ar_deduction_to_line(
                ar_deduction=justification,
                predicate_to_id=predicate_to_id,
                proven_predicate=proven_predicate,
            )
        case JustificationType.CIRCLE_MERGE:
            return _write_direct_justification_line(
                premises=(justification.circle.justification,),
                reason="Circle merge",
                proven_predicate=proven_predicate,
            )
        case JustificationType.LINE_MERGE:
            return _write_direct_justification_line(
                premises=(justification.direct_justification,),
                reason="Line merge",
                proven_predicate=proven_predicate,
            )
        case JustificationType.DIRECT_CONSEQUENCE:
            return _write_direct_justification_line(
                premises=justification.premises,
                reason="Direct consequence",
                proven_predicate=proven_predicate,
            )
    raise NotImplementedError(
        f"Writing proof for justification {justification} is not implemented"
    )


def _rule_application_to_line(
    proven_predicate: PredicateInProof,
    rule_application: RuleApplication,
    predicate_to_id: dict[Predicate, PredicateInProof],
) -> str:
    premises_in_proof = [
        predicate_to_id[premise] for premise in rule_application.premises
    ]
    premises_txt = ", ".join(
        f"{_predicate_str(premise_in_proof)}" for premise_in_proof in premises_in_proof
    )
    rule_str = rule_application.rule.fullname
    return f"{premises_txt} =({rule_str})> {_predicate_str(proven_predicate)}"


def _ar_deduction_to_line(
    ar_deduction: ARDeduction,
    predicate_to_id: dict[Predicate, PredicateInProof],
    proven_predicate: PredicateInProof,
) -> str:
    if ar_deduction.ar_premises is None:
        raise ValueError(f"AR deduction has no premises: {ar_deduction}")
    premises_in_proof = [
        predicate_to_id[premise.predicate] for premise in ar_deduction.ar_premises
    ]
    premises_txt = ", ".join(
        f"{_predicate_str(premise_in_proof)}" for premise_in_proof in premises_in_proof
    )
    return f"{premises_txt} =({ar_deduction.dependency_type.value})> {_predicate_str(proven_predicate)}"


def _write_direct_justification_line(
    premises: tuple[Predicate, ...],
    reason: str,
    proven_predicate: PredicateInProof,
) -> str:
    premises_txt = ", ".join(str(premise) for premise in premises)
    return f"{premises_txt} =({reason})> {_predicate_str(proven_predicate)}"


def _predicate_str(proven_predicate: PredicateInProof) -> str:
    return f"{str(proven_predicate.predicate)} [{proven_predicate.id}]"


class CarryOver:
    def __init__(self, carry_over: dict[str, Fraction] | None = None):
        if carry_over is None:
            carry_over = defaultdict(lambda: Fraction(0))
        self.carry_over = carry_over

    def update(self, ar_premise: ARPremise) -> CarryOver:
        new_carry_over = self.carry_over.copy()
        for table_column, column_value in ar_premise.coefficient.lhs_terms.items():
            new_carry_over[table_column] += ar_premise.coefficient.coeff * column_value
        return CarryOver(carry_over=new_carry_over)

    def __str__(self) -> str:
        return _pretty_fraction_dict(self.carry_over)


def _pretty_fraction_dict(fraction_dict: dict[str, Fraction]) -> str:
    content = ", ".join(
        f"{_pretty_column_name(column)}: {_fraction_str(value)}"
        for column, value in fraction_dict.items()
        if value != Fraction(0)
    )
    return f"{{{content}}}"


def _fraction_str(fraction: Fraction) -> str:
    return (
        f"{fraction.numerator}/{fraction.denominator}"
        if fraction.denominator != 1
        else str(fraction.numerator)
    )


def _pretty_column_name(column_name: str) -> str:
    return (
        column_name.replace("-", "")
        .replace("|", "")
        .replace("(", "")
        .replace(")", "")
        .replace("∠", "")
        .upper()
    ).replace("^2", "²")


def _write_ar_deduction_appendix(step_id: int, ar_deduction: ARDeduction) -> list[str]:
    if ar_deduction.ar_premises is None:
        raise ValueError(f"AR deduction has no premises: {ar_deduction}")

    ar_lines: list[str] = [
        f"{step_id:03d}. | {ar_deduction.ar_reason.value} to prove {ar_deduction.predicate}:"
    ]

    prefix = "| "
    carry_over = CarryOver()
    for premise in ar_deduction.ar_premises:
        new_carry_over = carry_over.update(premise)
        pretty_lhs = _pretty_fraction_dict(premise.coefficient.lhs_terms)
        ar_lines.append(
            f"{prefix}Premise {premise.predicate} gives a linear equation with coefficients {pretty_lhs}"
        )
        if carry_over.carry_over:
            ar_lines.append(
                f"{prefix}Then {carry_over} + {premise.coefficient.coeff} x {pretty_lhs}"
            )
            ar_lines.append(f"{prefix}   = {new_carry_over}")
        carry_over = new_carry_over

    ar_lines.append(
        f"We are left with a linear equation with coefficients {carry_over} that gives {ar_deduction.predicate}"
    )

    return ar_lines



from __future__ import annotations

from fractions import Fraction
from typing import TYPE_CHECKING, Literal, Optional

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Segment
from newclid.tools import fraction_to_len, get_quotient, str_to_fraction

if TYPE_CHECKING:
    from newclid.proof_state import ProofState


class ConstantLength(PredicateInterface):
    """lconst A B L -
    Represent that the length of segment AB is L

    L should be given as a float.
    """

    predicate_type: Literal[PredicateType.CONSTANT_LENGTH] = (
        PredicateType.CONSTANT_LENGTH
    )
    segment: Segment
    length: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, length = args
        if a == b:
            return None
        a, b = sorted((a, b))
        return (a, b, PredicateArgument(fraction_to_len(str_to_fraction(length))))

    def check_numerical(self) -> bool:
        a, b = self.segment
        actual_length = a.num.distance(b.num)
        return close_enough(actual_length, float(self.length))

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        a, b = self.segment
        return (
            PredicateArgument(a.name),
            PredicateArgument(b.name),
            PredicateArgument(fraction_to_len(self.length)),
        )

    def __str__(self) -> str:
        a, b = self.segment
        return f"{a}{b} = {fraction_to_len(self.length)}"


class LCompute(PredicateInterface):
    """lcompute A B"""

    predicate_type: Literal[PredicateType.L_COMPUTE] = PredicateType.L_COMPUTE
    segment: Segment

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> Optional[tuple[PredicateArgument, ...]]:
        a, b = args
        if a == b:
            return None
        return tuple(sorted((a, b)))

    def check_numerical(self) -> bool:
        return True  # Can always compute length

    def check(self, proof_state: ProofState) -> bool:
        lconst = lconst_from_lcompute(self)
        return proof_state.check(lconst)

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        return tuple(PredicateArgument(p.name) for p in self.segment)

    def __str__(self) -> str:
        return str(lconst_from_lcompute(self))


def lconst_from_lcompute(lcompute: LCompute) -> ConstantLength:
    a, b = lcompute.segment
    length = get_quotient(a.num.distance(b.num))
    return ConstantLength(segment=(a, b), length=length)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterable, Optional, Sequence, Union

import numpy as np
from numpy.random import Generator
from pydantic import BaseModel

from newclid.jgex.errors import (
    InvalidIntersectError,
    JGEXConstructionError,
    PointTooCloseError,
)
from newclid.numerical import close_enough, nearly_zero, sign

LENGTH_UNIT = 1.0
"""Base unit for lengths. Here to avoid numerical precision issues if dealing with too little numbers."""


class JGEXPoint(BaseModel):
    """Numerical point."""

    x: float
    y: float

    def __add__(self, p: JGEXPoint) -> JGEXPoint:
        return JGEXPoint(x=self.x + p.x, y=self.y + p.y)

    def __sub__(self, p: JGEXPoint) -> JGEXPoint:
        return JGEXPoint(x=self.x - p.x, y=self.y - p.y)

    def __mul__(self, f: Any) -> JGEXPoint:
        f = float(f)
        return JGEXPoint(x=self.x * f, y=self.y * f)

    def __rmul__(self, f: Any) -> JGEXPoint:
        return self * float(f)

    def __truediv__(self, f: Any) -> JGEXPoint:
        f = float(f)
        return JGEXPoint(x=self.x / f, y=self.y / f)

    def __str__(self) -> str:
        return "PointNum({},{})".format(self.x, self.y)

    def __abs__(self) -> str:
        return np.sqrt(self.dot(self))

    def angle(self) -> float:
        return np.arctan2(self.y, self.x)

    def close_enough(self, point: JGEXPoint) -> bool:
        return close_enough(self.x, point.x) and close_enough(self.y, point.y)

    def distance(self, p: Union[JGEXPoint, JGEXLine, JGEXCircle]) -> float:
        return np.sqrt(self.distance2(p))

    def distance2(self, p: Union[JGEXPoint, JGEXLine, JGEXCircle]) -> float:
        if isinstance(p, JGEXLine):
            return p.distance(self)
        if isinstance(p, JGEXCircle):
            return (p.radius - self.distance(p.center)) ** 2
        dx = self.x - p.x
        dy = self.y - p.y
        dx2 = dx * dx
        dy2 = dy * dy
        dx2 = 0.0 if nearly_zero(dx2) else dx2
        dy2 = 0.0 if nearly_zero(dy2) else dy2
        return dx2 + dy2

    def rot90(self) -> JGEXPoint:
        return JGEXPoint(x=-self.y, y=self.x)

    def rotatea(self, ang: Any) -> JGEXPoint:
        sinb, cosb = np.sin(ang), np.cos(ang)
        return self.rotate(sinb, cosb)

    def rotate(self, sinb: Any, cosb: Any) -> JGEXPoint:
        x, y = self.x, self.y
        return JGEXPoint(x=x * cosb - y * sinb, y=x * sinb + y * cosb)

    def flip(self) -> JGEXPoint:
        return JGEXPoint(x=-self.x, y=self.y)

    def perpendicular_line(self, line: JGEXLine) -> JGEXLine:
        return line.perpendicular_line(self)

    def foot(self, line: Union[JGEXLine, JGEXCircle]) -> JGEXPoint:
        if isinstance(line, JGEXLine):
            perpendicular_line = line.perpendicular_line(self)
            return line_line_intersection(perpendicular_line, line)[0]
        else:
            c, r = line.center, line.radius
            return c + (self - c) * r / self.distance(c)

    def parallel_line(self, line: JGEXLine) -> JGEXLine:
        return line.parallel_line(self)

    def dot(self, other: JGEXPoint) -> float:
        res = self.x * other.x + self.y * other.y
        return 0 if nearly_zero(res) else res

    @classmethod
    def deduplicate(cls, points: Iterable[JGEXPoint]) -> list[JGEXPoint]:
        res: list[JGEXPoint] = []
        for p in points:
            if all(not r.close_enough(p) for r in res):
                res.append(p)
        return res

    def intersect(self, obj: JGEXGeometry) -> list[JGEXPoint]:
        raise NotImplementedError()

    def __repr__(self) -> str:
        return f"JGEXPoint({self.x}, {self.y})"


class JGEXGeometry(ABC):
    @abstractmethod
    def sample_within(
        self, points: Sequence[JGEXPoint], *, trials: int = 5, rng: Generator
    ) -> JGEXPoint: ...


class JGEXLine(JGEXGeometry):
    """Numerical line."""

    def __init__(
        self,
        p1: Optional["JGEXPoint"] = None,
        p2: Optional["JGEXPoint"] = None,
        coefficients: Optional[tuple[float, float, float]] = None,
    ):
        a, b, c = None, None, None
        if coefficients:
            a, b, c = coefficients
        elif p1 and p2:
            if p1.close_enough(p2):
                raise PointTooCloseError(
                    f"Not able to determine the line by points {p1} and {p2} as they are too close (distance: {p1.distance(p2)})"
                )
            a, b, c = (
                p1.y - p2.y,
                p2.x - p1.x,
                p1.x * p2.y - p2.x * p1.y,
            )
        assert a is not None and b is not None and c is not None

        # Make sure a is always positive (or always negative for that matter)
        # With a == 0, Assuming a = +epsilon > 0
        # Then b such that ax + by = 0 with y>0 should be negative.
        sa = sign(a)
        sb = sign(b)
        if sa == -1 or sa == 0 and sb == 1:
            a, b, c = -a, -b, -c
        if sa == 0:
            a = 0.0
        if sb == 0:
            b = 0.0
        if nearly_zero(c):
            c = 0.0

        d: float = np.sqrt(a**2 + b**2)
        self.coefficients: tuple[float, float, float] = a / d, b / d, c / d

    def __call__(self, p: JGEXPoint) -> float:
        a, b, c = self.coefficients
        res = p.x * a + p.y * b + c
        return 0 if nearly_zero(res) else res

    def parallel_line(self, p: JGEXPoint) -> JGEXLine:
        a, b, _ = self.coefficients
        return JGEXLine(coefficients=(a, b, -a * p.x - b * p.y))

    def perpendicular_line(self, p: JGEXPoint) -> JGEXLine:
        a, b, _ = self.coefficients
        return JGEXLine(p, p + JGEXPoint(x=a, y=b))

    def distance(self, p: JGEXPoint) -> float:
        return abs(self(p))

    def is_parallel(self, other: JGEXLine) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return close_enough(a * y, b * x)

    def is_perp(self, other: JGEXLine) -> bool:
        a, b, _ = self.coefficients
        x, y, _ = other.coefficients
        return close_enough(a * x, -b * y)

    def is_same(self, other: JGEXLine) -> bool:
        a, b, c = self.coefficients
        x, y, z = other.coefficients
        return close_enough(a * y, b * x) and close_enough(b * z, c * y)

    def point_at(
        self, x: Optional[float] = None, y: Optional[float] = None
    ) -> Optional[JGEXPoint]:
        """Infer the point on the line by its x and/or y coordinate(s)"""
        a, b, c = self.coefficients
        # ax + by + c = 0
        if x is None and y is not None:
            if not close_enough(a, 0):
                return JGEXPoint(x=(-c - b * y) / a, y=y)
            else:
                return None
        elif x is not None and y is None:
            if not close_enough(b, 0):
                return JGEXPoint(x=x, y=(-c - a * x) / b)
            else:
                return None
        elif x is not None and y is not None:
            if close_enough(a * x + b * y, -c):
                return JGEXPoint(x=x, y=y)
        return None

    def diff_side(self, p1: JGEXPoint, p2: JGEXPoint) -> bool:
        d1 = self(p1)
        d2 = self(p2)
        if nearly_zero(d1) or nearly_zero(d2):
            return False
        return d1 * d2 < 0

    def same_side(self, p1: JGEXPoint, p2: JGEXPoint) -> bool:
        d1 = self(p1)
        d2 = self(p2)
        if close_enough(d1, 0) or close_enough(d2, 0):
            return False
        return d1 * d2 > 0

    def sample_within(
        self, points: Sequence[JGEXPoint], *, trials: int = 5, rng: Generator
    ) -> JGEXPoint:
        """Sample a point on the line within the boundary of existing points."""
        center: JGEXPoint = sum(points, JGEXPoint(x=0.0, y=0.0)) / len(points)  # type: ignore
        radius = max([p.distance(center) for p in points])

        if close_enough(center.distance(self), radius):
            center = center.foot(self)
        a, b = line_circle_intersection(
            self, JGEXCircle(center.foot(self), radius), expected_points=2
        )
        result = None
        best = -1.0
        for _ in range(trials):
            rand = rng.uniform(0.0, 1.0)
            x = a + (b - a) * rand
            mind = min([x.distance(p) for p in points])
            if mind > best:
                best = mind
                result = x

        if result is None:
            raise JGEXConstructionError(
                "Could not sample a point on line within the boundary of points"
            )
        return result

    def angle(self) -> float:
        if nearly_zero(self.coefficients[1]):
            return np.pi / 2
        phead = self.point_at(x=1)
        ptail = self.point_at(x=0)
        if phead is None or ptail is None:
            raise ValueError("Cannot compute the angle if no head or tail.")
        angle = (phead - ptail).angle() % np.pi
        if close_enough(angle, np.pi):
            return 0.0
        return angle

    def angle_to(self, other: JGEXLine) -> float:
        return (self.angle() - other.angle()) % np.pi

    def __repr__(self) -> str:
        return f"JGEXLine({self.coefficients[0]}, {self.coefficients[1]}, {self.coefficients[2]})"


class JGEXCircle(JGEXGeometry):
    """Numerical circle."""

    def __init__(
        self,
        center: Optional[JGEXPoint] = None,
        radius: Optional[float] = None,
        p1: Optional[JGEXPoint] = None,
        p2: Optional[JGEXPoint] = None,
        p3: Optional[JGEXPoint] = None,
    ):
        if center:
            self.center = center
        else:
            if not (p1 and p2 and p3):
                raise ValueError("Circle without center need p1 p2 p3")
            l12 = perpendicular_bisector(p1, p2)
            l23 = perpendicular_bisector(p2, p3)
            (self.center,) = line_line_intersection(l12, l23)

        self.a, self.b = self.center.x, self.center.y
        self.points = [p for p in (p1, p2, p3) if p is not None]

        if not radius:
            p = p1 or p2 or p3
            if p is None:
                raise ValueError("Circle needs radius or p1 or p2 or p3")
            self.r2: float = (self.a - p.x) ** 2 + (self.b - p.y) ** 2
            self.radius: float = np.sqrt(self.r2)
        else:
            self.radius = radius
            self.r2 = radius * radius

    def sample_within(
        self, points: Sequence[JGEXPoint], *, trials: int = 5, rng: Generator
    ) -> JGEXPoint:
        """Sample a point on the circle."""
        random_angle = rng.uniform(0.0, 2.0) * np.pi
        sampled_point = self.center + self.radius * JGEXPoint(
            x=np.cos(random_angle), y=np.sin(random_angle)
        )
        return sampled_point

    def __repr__(self) -> str:
        return f"JGEXCircle(c={self.center}, r={self.radius}, points={self.points}])"


def perpendicular_bisector(p1: JGEXPoint, p2: JGEXPoint) -> JGEXLine:
    midpoint = (p1 + p2) * 0.5
    return JGEXLine(midpoint, midpoint + JGEXPoint(x=p2.y - p1.y, y=p1.x - p2.x))


def solve_quad(a: float, b: float, c: float) -> tuple[float, ...]:
    """Solve a x^2 + bx + c = 0."""
    if nearly_zero(a):
        return () if nearly_zero(b) else (-c / b,)
    a = 2 * a
    d = b * b - 2 * a * c
    sd = sign(d)
    if sd == -1:
        return ()
    if sd == 0:
        d = 0.0
    y = np.sqrt(d)
    if nearly_zero(y):
        return (-b / a,)
    return (-b - y) / a, (-b + y) / a


def intersect(a: JGEXGeometry, b: JGEXGeometry):
    if isinstance(a, JGEXCircle):
        if isinstance(b, JGEXCircle):
            return circle_circle_intersection(a, b)
        if isinstance(b, JGEXLine):
            return line_circle_intersection(b, a)
    if isinstance(a, JGEXLine):
        if isinstance(b, JGEXCircle):
            return line_circle_intersection(a, b)
        if isinstance(b, JGEXLine):
            return line_line_intersection(a, b, ensure_point=False)
    raise NotImplementedError


def circle_circle_intersection(
    c1: JGEXCircle, c2: JGEXCircle, expected_points: int | None = None
) -> tuple[JGEXPoint, ...]:
    """Returns a pair of Points as intersections of c1 and c2."""
    # circle 1: (x0, y0), radius r0
    # circle 2: (x1, y1), radius r1
    x0, y0, r0 = c1.a, c1.b, c1.radius
    x1, y1, r1 = c2.a, c2.b, c2.radius

    d: float = (x1 - x0) ** 2 + (y1 - y0) ** 2
    if nearly_zero(d):
        raise InvalidIntersectError(
            f"Circles {c1} and {c2} have the same center and thus cannot intersect"
        )
    d = np.sqrt(d)

    if not (r0 + r1 >= d and abs(r0 - r1) <= d):
        raise InvalidIntersectError(f"Circles {c1} and {c2} do intersect")

    a = (r0**2 - r1**2 + d**2) / (2 * d)
    h = r0**2 - a**2
    biu: JGEXPoint = (c2.center - c1.center) / d
    qiu = biu.rot90()
    p = c1.center + a * biu
    if nearly_zero(h):
        if expected_points is not None and expected_points != 1:
            raise InvalidIntersectError(
                f"Expected {expected_points} intersections between circles {c1} and {c2}, but they are tangent"
            )
        return (p,)

    if expected_points is not None and expected_points != 2:
        raise InvalidIntersectError(
            f"Expected {expected_points} intersections between circles, but they are crossing"
        )

    qiu = np.sqrt(h) * qiu

    return (p + qiu, p - qiu)


def line_circle_intersection(
    line: JGEXLine, circle: JGEXCircle, expected_points: int = 1
) -> tuple[JGEXPoint, ...]:
    a, b, c = line.coefficients
    r = circle.radius
    center = circle.center
    p, q = center.x, center.y
    if nearly_zero(b):
        x = -c / a
        x_p = x - p
        x_p2 = x_p * x_p
        points = tuple(
            JGEXPoint(x=x, y=y1) for y1 in solve_quad(1, -2 * q, q * q + x_p2 - r * r)
        )
    elif close_enough(a, 0):
        y = -c / b
        y_q = y - q
        y_q2 = y_q * y_q
        points = tuple(
            JGEXPoint(x=x1, y=y) for x1 in solve_quad(1, -2 * p, p * p + y_q2 - r * r)
        )

    else:
        c_ap = c + a * p
        a2 = a * a
        points = tuple(
            JGEXPoint(x=-(b * y1 + c) / a, y=y1)
            for y1 in solve_quad(
                a2 + b * b, 2 * (b * c_ap - a2 * q), c_ap * c_ap + a2 * (q * q - r * r)
            )
        )
    if len(points) < expected_points:
        raise InvalidIntersectError(
            f"Expected at least {expected_points} intersections between line and circle, got {len(points)}"
        )
    return points


def line_line_intersection(
    line_1: JGEXLine, line_2: JGEXLine, ensure_point: bool = True
) -> tuple[JGEXPoint, ...]:
    a1, b1, c1 = line_1.coefficients
    a2, b2, c2 = line_2.coefficients
    # a1x + b1y + c1 = 0
    # a2x + b2y + c2 = 0
    d = a1 * b2 - a2 * b1
    if nearly_zero(d):
        if ensure_point:
            raise InvalidIntersectError(
                f"Expected 1 intersection between lines, got {d}"
            )
        return ()
    return (JGEXPoint(x=(c2 * b1 - c1 * b2) / d, y=(c1 * a2 - c2 * a1) / d),)


def reduce_intersection(
    objs: Sequence[JGEXGeometry],
    existing_points: Sequence[JGEXPoint],
    rng: Generator,
) -> list[JGEXPoint]:
    """Reduce intersecting objects into one new point."""
    if all(isinstance(o, JGEXPoint) for o in objs):
        return list(objs)  # type: ignore

    if len(objs) == 1:
        obj = objs[0]
        return [obj.sample_within(existing_points, rng=rng)]

    if len(objs) == 2:
        u, v = objs
        intersections: tuple[JGEXPoint, ...] = intersect(u, v)
        shuffled_intersections = np.array(intersections)  # pyright: ignore
        rng.shuffle(shuffled_intersections)  # pyright: ignore
        shuffled_intersections: tuple[JGEXPoint, ...] = tuple(shuffled_intersections)  # type: ignore
        for p in shuffled_intersections:
            is_valid = all(not p.close_enough(x) for x in existing_points)
            if is_valid:
                return [p]
        raise InvalidIntersectError(
            f"Could not reduce {objs} to a single new point, intersections: {intersections}, existing points: {existing_points}"
        )

    raise NotImplementedError



from typing import NewType

import numpy as np
from sympy.core.backend import Symbol as SympySymbol  # type: ignore

from newclid.deductors.sympy_ar.ar_table import ARTable
from newclid.numerical import close_enough
from newclid.numerical.geometries import LineNum
from newclid.symbols.lines_registry import LineSymbol

LineSympySymbol = NewType("LineSympySymbol", SympySymbol)  # type: ignore


class AnglesTable:
    def __init__(self, ar_table: ARTable) -> None:
        self.inner_table = ar_table

        self.expected_lines: dict[LineSympySymbol, LineNum] = {}
        self.expected_aconsts: dict[tuple[LineSympySymbol, LineSympySymbol], float] = {}
        self.expected_parallels: set[tuple[LineSympySymbol, LineSympySymbol]] = set()
        self.expected_perpendiculars: set[tuple[LineSympySymbol, LineSympySymbol]] = (
            set()
        )
        self.expected_eqangles: set[
            tuple[LineSympySymbol, LineSympySymbol, LineSympySymbol, LineSympySymbol]
        ] = set()

        self.newclid_to_sympy_symbol: dict[LineSymbol, LineSympySymbol] = {}
        self.sympy_to_newclid_symbol: dict[LineSympySymbol, LineSymbol] = {}

    def line_symbol(self, newclid_symbol: LineSymbol) -> SympySymbol:
        if newclid_symbol not in self.newclid_to_sympy_symbol:
            angle_symbol = LineSympySymbol(SympySymbol(str(newclid_symbol)))
            self.newclid_to_sympy_symbol[newclid_symbol] = angle_symbol
            self.sympy_to_newclid_symbol[angle_symbol] = newclid_symbol
            newclid_symbol = self.sympy_to_newclid_symbol[angle_symbol]
            self._add_angle_absolute(angle_symbol, num=newclid_symbol.num)
        return self.newclid_to_sympy_symbol[newclid_symbol]

    def _add_angle_absolute(self, angle: LineSympySymbol, num: LineNum):
        for other_angle, other_line_num in self.expected_lines.items():
            aconst = (angle, other_angle)
            aconst_value = num.angle_to(other_line_num)

            if num.is_parallel(other_line_num):
                self.expected_parallels.add((angle, other_angle))
                continue

            if num.is_perp(other_line_num):
                self.expected_perpendiculars.add((angle, other_angle))

            reversed_aconst = (other_angle, angle)
            reversed_aconst_value = (np.pi - aconst_value) % np.pi

            # Add the eqangle possibilities for both directions.
            for other_aconst, other_aconst_value in self.expected_aconsts.items():
                if close_enough(aconst_value, other_aconst_value):
                    self.expected_eqangles.add((*aconst, *other_aconst))
                if close_enough(reversed_aconst_value, other_aconst_value):
                    self.expected_eqangles.add((*reversed_aconst, *other_aconst))

            self.expected_aconsts[aconst] = aconst_value
            self.expected_aconsts[reversed_aconst] = reversed_aconst_value

        self.expected_lines[angle] = num



"""Main loop of interactions between the agent and the proof state."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

from pydantic import BaseModel

from newclid.agent import AgentStats
from newclid.rule import Rule

if TYPE_CHECKING:
    from newclid.agent.agents_interface import DeductiveAgent
    from newclid.proof_state import ProofState


class RunInfos(BaseModel):
    runtime: float
    success: bool
    steps: int
    success_per_goal: dict[str, bool]
    agent_stats: AgentStats | None = None


def run_loop(
    deductive_agent: DeductiveAgent, proof: "ProofState", rules: list[Rule]
) -> RunInfos:
    """Run DeductiveAgent until saturation or goal found."""
    for goal in proof.goals:
        if not proof.check_numerical(goal):
            raise ValueError("%s fails numerical check", goal)

    t0 = time.time()
    step = 0
    running = True
    while running:
        running = deductive_agent.step(proof=proof, rules=rules)
        step += 1

    runtime = time.time() - t0
    success_per_goal: dict[str, bool] = {}
    for goal in proof.goals:
        success_per_goal[f"{goal} succeeded"] = proof.check(goal)

    return RunInfos(
        runtime=runtime,
        success=all(success_per_goal.values()),
        success_per_goal=success_per_goal,
        steps=step,
        agent_stats=deductive_agent.get_stats(),
    )



import numpy as np

from newclid.jgex.geometries import JGEXPoint


def ang_of(tail: JGEXPoint, head: JGEXPoint) -> float:
    vector = head - tail
    arctan = np.arctan2(vector.y, vector.x) % (2 * np.pi)
    return arctan


def ang_between(tail: JGEXPoint, head1: JGEXPoint, head2: JGEXPoint) -> float:
    """
    the slid angle from tail->head1 to tail->head2 controlled between [-np.pi, np.pi)
    """
    ang1 = ang_of(tail, head1)
    ang2 = ang_of(tail, head2)
    diff = ang2 - ang1
    return diff % (2 * np.pi)



from __future__ import annotations

import logging

from numpy.random import Generator as RngGenerator
from pydantic import BaseModel

from newclid.jgex.clause import JGEXClause, JGEXConstruction, is_numerical_argument
from newclid.jgex.definition import (
    JGEXDefinition,
    JGEXVariableName,
    SketchCall,
    mapping_from_construction,
)
from newclid.jgex.distances import (
    ensure_not_too_close_numerical,
    ensure_not_too_far_numerical,
)
from newclid.jgex.errors import JGEXConstructionError
from newclid.jgex.formulation import JGEXFormulation
from newclid.jgex.geometries import JGEXGeometry, JGEXPoint, reduce_intersection
from newclid.jgex.sketch import sketch
from newclid.numerical.geometries import PointNum
from newclid.predicate_types import PredicateArgument
from newclid.predicates import NUMERICAL_PREDICATES
from newclid.predicates._index import PredicateType
from newclid.problem import (
    PredicateConstruction,
    ProblemSetup,
    rename_predicate_construction,
)
from newclid.symbols.points_registry import Point
from newclid.tools import atomize

LOGGER = logging.getLogger(__name__)


def build_newclid_problem(
    problem: JGEXFormulation,
    defs: dict[str, JGEXDefinition],
    rng: RngGenerator,
    max_attempts_per_clause: int,
    include_auxiliary_clauses: bool,
) -> tuple[ProblemSetup, dict[JGEXClause, JGEXClauseConsequences]]:
    """Build a JGEX problem by finding coordinates for the points given the clauses constraints."""
    LOGGER.debug(f"Building JGEX problem '{problem.name}': {problem}")

    nc_problem = ProblemSetup(
        name=problem.name,
        points=(),
        assumptions=(),
        goals=(),
    )
    clauses_consequences: dict[JGEXClause, JGEXClauseConsequences] = {}

    clauses = problem.clauses if include_auxiliary_clauses else problem.setup_clauses
    for clause in clauses:
        nc_problem, clause_consequences = add_clause_to_problem(
            problem=nc_problem,
            clause=clause,
            defs=defs,
            rng=rng,
            max_attempts=max_attempts_per_clause,
        )
        clauses_consequences[clause] = clause_consequences

    nc_problem = nc_problem.with_new(new_goals=problem.goals)
    return nc_problem, clauses_consequences


def add_clause_to_problem(
    problem: ProblemSetup,
    clause: JGEXClause,
    defs: dict[str, JGEXDefinition],
    rng: RngGenerator,
    max_attempts: int,
) -> tuple[ProblemSetup, JGEXClauseConsequences]:
    """Add a clause to a problem, returning the new problem and the consequences of the clause.

    Raises:
        JGEXConstructionError: If the clause cannot be added to the problem after max_attempts.

    """
    points: dict[PredicateArgument, JGEXPoint] = {
        PredicateArgument(pt.name): JGEXPoint(x=pt.num.x, y=pt.num.y)
        for pt in problem.points
    }
    default_error_msg = "Did not even attempt to add a clause to the problem"
    last_error = JGEXConstructionError(default_error_msg)
    for attempt in range(max_attempts):
        try:
            clause_consequences = _add_jgex_clause(clause, points, defs, rng)
            return (
                problem.with_new(
                    new_points=_jgex_points_to_newclid_points(
                        clause_consequences.new_points
                    ),
                    new_assumptions=clause_consequences.to_assumptions(),
                ),
                clause_consequences,
            )
        except JGEXConstructionError as e:
            LOGGER.debug(
                f"Failed to add clause to problem {problem.name} (attempt {attempt}): {e}"
            )
            last_error = e
            continue
    raise JGEXConstructionError(
        f"Failed to add clause to problem {problem.name} after {max_attempts} attempts. Last error: {last_error}"
    )


def _jgex_points_to_newclid_points(
    points: dict[PredicateArgument, JGEXPoint],
) -> tuple[Point, ...]:
    return tuple(
        Point(name=PredicateArgument(pname), num=PointNum(x=p.x, y=p.y))
        for pname, p in points.items()
    )


class JGEXClauseConsequences(BaseModel):
    new_points: dict[PredicateArgument, JGEXPoint]
    construction_consequences: list[JGEXConstructionConsequences]
    numerical_requirements: list[PredicateConstruction]

    def to_assumptions(self) -> tuple[PredicateConstruction, ...]:
        return tuple(
            predicate_construction
            for construction_consequence in self.construction_consequences
            for predicate_construction in construction_consequence.to_assumptions()
        )


def rename_points_in_clause_consequences(
    clause_consequences: JGEXClauseConsequences,
    mapping: dict[PredicateArgument, PredicateArgument],
) -> JGEXClauseConsequences:
    return JGEXClauseConsequences(
        new_points={
            mapping[pname]: p for pname, p in clause_consequences.new_points.items()
        },
        construction_consequences=[
            rename_points_in_construction_consequences(
                construction_consequence, mapping
            )
            for construction_consequence in clause_consequences.construction_consequences
        ],
        numerical_requirements=[
            rename_predicate_construction(predicate_construction, mapping)
            for predicate_construction in clause_consequences.numerical_requirements
        ],
    )


def _add_jgex_clause(
    clause: JGEXClause,
    existing_points: dict[PredicateArgument, JGEXPoint],
    defs: dict[str, JGEXDefinition],
    rng: RngGenerator,
) -> JGEXClauseConsequences:
    """Return a new problem with a new clause of construction, e.g. a new excenter."""
    sketches_calls: list[SketchCall] = []
    constructions: list[JGEXConstructionConsequences] = []
    numerical_requirements: list[PredicateConstruction] = []
    for construction in clause.constructions:
        construction_definition = defs.get(construction.name)
        if construction_definition is None:
            if construction.name in {
                num_predicate_type.value
                for num_predicate_type in NUMERICAL_PREDICATES.keys()
            }:
                numerical_requirements.append(
                    PredicateConstruction.from_tuple(
                        (construction.name, *construction.args)
                    )
                )
                continue
            raise JGEXConstructionError(
                f"Construction {construction} has no definition"
            )

        clause_construction, mapping = _add_construction(
            construction=construction,
            construction_definition=construction_definition,
            existing_points=existing_points,
            clause=clause,
        )
        constructions.append(clause_construction)
        sketches_calls.extend(
            SketchCall.from_construction(sketch, mapping)
            for sketch in construction_definition.sketches
        )
    new_points = _add_clause_points(clause, existing_points, sketches_calls, rng)
    return JGEXClauseConsequences(
        construction_consequences=constructions,
        numerical_requirements=numerical_requirements,
        new_points=new_points,
    )


class JGEXConstructionConsequences(BaseModel):
    required_predicates: list[PredicateConstruction]
    added_predicates: list[PredicateConstruction]

    def to_assumptions(self) -> tuple[PredicateConstruction, ...]:
        return tuple(self.required_predicates + self.added_predicates)


def rename_points_in_construction_consequences(
    construction_consequences: JGEXConstructionConsequences,
    mapping: dict[PredicateArgument, PredicateArgument],
) -> JGEXConstructionConsequences:
    return JGEXConstructionConsequences(
        required_predicates=[
            rename_predicate_construction(predicate_construction, mapping)
            for predicate_construction in construction_consequences.required_predicates
        ],
        added_predicates=[
            rename_predicate_construction(predicate_construction, mapping)
            for predicate_construction in construction_consequences.added_predicates
        ],
    )


def _add_construction(
    construction: JGEXConstruction,
    construction_definition: JGEXDefinition,
    existing_points: dict[PredicateArgument, JGEXPoint],
    clause: JGEXClause,
) -> tuple[
    JGEXConstructionConsequences,
    dict[JGEXVariableName, PredicateArgument],
]:
    added: list[PredicateConstruction] = []
    mapping = mapping_from_construction(construction, construction_definition)

    # Ensure that all input points exist
    for input_point in construction_definition.input_points:
        required_point = mapping.get(input_point)
        if required_point is None:
            raise ValueError(
                f"Definition {construction_definition.name.value} has input point {input_point} that is missing from mapping {mapping}. "
                f"This is probably just because the definition {construction_definition.name.name} error itself is wrong and should be fixed."
            )
        if required_point not in existing_points:
            raise ValueError(
                f"Construction '{construction}' with mapping {mapping}"
                f" requires input point '{required_point}' that does not exist yet: {list(existing_points.keys())}."
            )

    required: list[PredicateConstruction] = []
    for requirement in construction_definition.requirements.constructions:
        if len(requirement.args) == 0:
            continue
        required.append(
            _jgex_construction_to_predicate_construction(requirement, mapping)
        )

    for clause in construction_definition.clauses:
        for construction in clause.constructions:
            added.append(
                _jgex_construction_to_predicate_construction(construction, mapping)
            )

    construction_consequences = JGEXConstructionConsequences(
        required_predicates=required, added_predicates=added
    )
    return construction_consequences, mapping


def _jgex_construction_to_predicate_construction(
    construction: JGEXConstruction,
    mapping: dict[JGEXVariableName, PredicateArgument],
) -> PredicateConstruction:
    args: tuple[PredicateArgument, ...] = tuple(  # type: ignore
        mapping.get(arg, arg)  # type: ignore
        for arg in construction.args
    )
    return PredicateConstruction.from_predicate_type_and_args(
        PredicateType(construction.name), args
    )


def _add_clause_points(
    clause: JGEXClause,
    existing_points: dict[PredicateArgument, JGEXPoint],
    sketches_calls: list[SketchCall],
    rng: RngGenerator,
) -> dict[PredicateArgument, JGEXPoint]:
    new_fixed_points: dict[PredicateArgument, JGEXPoint | None] = {}
    for point_value in clause.points:
        if "@" in point_value:
            name, pos = atomize(point_value, "@")
            x, y = atomize(pos, "_")
            new_fixed_points[PredicateArgument(name)] = JGEXPoint(
                x=float(x), y=float(y)
            )
        else:
            new_fixed_points[PredicateArgument(point_value)] = None

    for pname in new_fixed_points:
        if pname in existing_points:
            raise JGEXConstructionError(
                f"Clause {clause} cannot be added because it creates point {pname} that already exists. "
                f"Current points are {list(existing_points.keys())}."
            )

    new_sketch_points = _sketch_new_numerical_points(
        sketches_calls, existing_points, rng
    )
    for (pname, fixed_point), sketch_point in zip(
        new_fixed_points.items(), new_sketch_points
    ):
        if fixed_point is None:
            new_fixed_points[pname] = sketch_point

    new_points = {pname: p for pname, p in new_fixed_points.items() if p is not None}
    _ensure_numerical_points_are_valid(new_points, existing_points)
    return new_points


def _sketch_new_numerical_points(
    sketches_calls: list[SketchCall],
    existing_points: dict[PredicateArgument, JGEXPoint],
    rng: RngGenerator,
) -> list[JGEXPoint]:
    to_be_intersected: list[JGEXGeometry] = []
    for sketch_call in sketches_calls:
        args: list[JGEXPoint | PredicateArgument] = []
        for arg in sketch_call.args:
            if is_numerical_argument(arg):
                # Handle numerical arguments
                args.append(PredicateArgument(arg))
            else:
                # Handle point names
                args.append(existing_points[arg])
        to_be_intersected += sketch(name=sketch_call.name, args=tuple(args), rng=rng)
    return reduce_intersection(
        to_be_intersected, list(existing_points.values()), rng=rng
    )


def _ensure_numerical_points_are_valid(
    new_points: dict[PredicateArgument, JGEXPoint],
    existing_points: dict[PredicateArgument, JGEXPoint],
) -> None:
    ensure_not_too_close_numerical(new_points, existing_points)
    ensure_not_too_far_numerical(new_points, existing_points)



from __future__ import annotations

from fractions import Fraction
from typing import Literal

from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.predicates._interface import PredicateInterface
from newclid.symbols.points_registry import Segment
from newclid.tools import fraction_to_len, str_to_fraction


class SquaredConstantLength(PredicateInterface):
    """l2const A B L -
    Represent that the square of the length of segment AB is L

    L should be given as a float.
    """

    predicate_type: Literal[PredicateType.SQUARED_CONSTANT_LENGTH] = (
        PredicateType.SQUARED_CONSTANT_LENGTH
    )
    segment: Segment
    square_length: Fraction

    @staticmethod
    def preparse(
        args: tuple[PredicateArgument, ...],
    ) -> tuple[PredicateArgument, ...] | None:
        a, b, slength = args
        if a == b:
            return None
        a, b = sorted((a, b))
        return (a, b, PredicateArgument(fraction_to_len(str_to_fraction(slength))))

    def check_numerical(self) -> bool:
        a, b = self.segment
        slength = self.square_length
        squared_length = (a.num - b.num).dot(a.num - b.num)
        return close_enough(squared_length, float(slength))

    def to_tokens(self) -> tuple[PredicateArgument, ...]:
        a, b = self.segment
        return (
            PredicateArgument(a.name),
            PredicateArgument(b.name),
            PredicateArgument(fraction_to_len(self.square_length)),
        )

    def __str__(self) -> str:
        a, b = self.segment
        return f"|{a}{b}|\u00b2 = {fraction_to_len(self.square_length)}"



from __future__ import annotations

import itertools
import logging
from fractions import Fraction
from typing import Iterator, TypeVar, cast

import numpy as np
import sympy as sp  # type: ignore

from newclid.deductors.sympy_ar.table_angles import AnglesTable
from newclid.deductors.sympy_ar.table_ratios import RatiosTable
from newclid.numerical import close_enough
from newclid.predicate_types import PredicateArgument
from newclid.predicates._index import PredicateType
from newclid.problem import PredicateConstruction
from newclid.symbols.lines_registry import LineSymbol
from newclid.symbols.points_registry import Segment
from newclid.symbols.symbols_registry import SymbolsRegistry
from newclid.tools import fraction_to_angle, fraction_to_ratio, get_quotient

TableType = TypeVar("TableType", bound=AnglesTable | RatiosTable)

LOGGER = logging.getLogger(__name__)


def coll_expressions(lines: list[LineSymbol], table: AnglesTable) -> list[sp.Expr]:
    expressions: list[sp.Expr] = []
    for line in lines:
        expression = cast(
            sp.Expr,
            table.line_symbol(line) - table.line_symbol(line),  # type: ignore
        )
        expressions.append(expression)
    return expressions


def parallel_expressions(
    l1: LineSymbol,
    l2: LineSymbol,
    table: AnglesTable,
) -> list[sp.Expr]:
    v1 = table.line_symbol(l1)
    v2 = table.line_symbol(l2)
    expression = cast(sp.Expr, v1 - v2)  # type: ignore
    return [expression]


def enumerate_parallels(
    table: AnglesTable, symbols: SymbolsRegistry
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for v1, v2 in table.expected_parallels.copy():
        para_expression: sp.Expr = cast(sp.Expr, v2 - v1)  # pyright: ignore
        if not table.inner_table.expr_delta(para_expression):
            continue

        l1 = table.sympy_to_newclid_symbol[v1]
        l2 = table.sympy_to_newclid_symbol[v2]

        l1_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l1)
        l2_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l2)

        for pt_0, pt_1 in itertools.combinations(l1_points, 2):
            for pt_2, pt_3 in itertools.combinations(l2_points, 2):
                args = tuple(
                    PredicateArgument(pt.name) for pt in (pt_0, pt_1, pt_2, pt_3)
                )
                para_construction = PredicateConstruction.from_predicate_type_and_args(
                    PredicateType.PARALLEL, args
                )
                yield para_construction, para_expression


def perpendicular_expressions(
    l1: LineSymbol,
    l2: LineSymbol,
    table: AnglesTable,
) -> list[sp.Expr]:
    v1 = table.line_symbol(l1)
    v2 = table.line_symbol(l2)
    expression = cast(sp.Expr, v1 - v2)  # type: ignore
    return [expression]


def enumerate_perpendiculars(
    table: AnglesTable, symbols: SymbolsRegistry
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for v1, v2 in table.expected_perpendiculars.copy():
        perp_expression: sp.Expr = cast(sp.Expr, v2 - v1)  # pyright: ignore
        if not table.inner_table.expr_delta(perp_expression):
            continue

        l1 = table.sympy_to_newclid_symbol[v1]
        l2 = table.sympy_to_newclid_symbol[v2]

        l1_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l1)
        l2_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l2)

        # This has to match the order that is in Perp.constant_expressions.
        for pt_0, pt_1 in itertools.combinations(l1_points, 2):
            for pt_2, pt_3 in itertools.combinations(l2_points, 2):
                args = tuple(
                    PredicateArgument(pt.name) for pt in (pt_0, pt_1, pt_2, pt_3)
                )
                perp_construction = PredicateConstruction.from_predicate_type_and_args(
                    PredicateType.PERPENDICULAR, args
                )
                yield perp_construction, perp_expression


def aconst_expression(
    l1: LineSymbol,
    l2: LineSymbol,
    angle: Fraction,
    table: AnglesTable,
) -> list[sp.Expr]:
    v1 = table.line_symbol(l1)
    v2 = table.line_symbol(l2)
    va = sp.Rational(angle.numerator, angle.denominator)
    expression = cast(sp.Expr, v2 - v1 - va)  # type: ignore
    return [expression]


def enumerate_aconsts(
    table: AnglesTable, symbols: SymbolsRegistry
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for (l1, l2), expected_aconst_value in table.expected_aconsts.copy().items():
        expression = cast(sp.Expr, l1 - l2)  # pyright: ignore
        subbed_in = table.inner_table.substitute_in_existing_expressions(expression)
        if not len(subbed_in.free_symbols) == 0:
            continue

        aconst_value = (-subbed_in) % 1  # pyright: ignore
        n, d = sp.fraction(aconst_value)  # pyright: ignore

        aconst_rad_value = float(np.pi * subbed_in % np.pi)  # pyright: ignore
        if not close_enough(aconst_rad_value, expected_aconst_value):
            LOGGER.debug(
                f"Angle {aconst_value} ({aconst_rad_value}) is the expected numerical value {expected_aconst_value} in between lines {(l1, l2)}"
            )

        if n == 0:
            continue

        l1_newclid = symbols.lines.representent(table.sympy_to_newclid_symbol[l1])
        l2_newclid = symbols.lines.representent(table.sympy_to_newclid_symbol[l2])

        for p1, p2 in itertools.combinations(l1_newclid.points, 2):
            for p3, p4 in itertools.combinations(l2_newclid.points, 2):
                p1, p2 = sorted((p1, p2), key=lambda x: x.name)
                p3, p4 = sorted((p3, p4), key=lambda x: x.name)
                fraction = PredicateArgument(
                    fraction_to_angle(Fraction(int(n), int(d)))
                )
                points = tuple(PredicateArgument(pt.name) for pt in (p1, p2, p3, p4))
                yield (
                    PredicateConstruction.from_predicate_type_and_args(
                        PredicateType.CONSTANT_ANGLE,
                        args=tuple(points) + (fraction,),
                    ),
                    expression,
                )


def eqangle_expressions(
    l1: LineSymbol,
    l2: LineSymbol,
    l3: LineSymbol,
    l4: LineSymbol,
    table: AnglesTable,
) -> list[sp.Expr]:
    v1 = table.line_symbol(l1)
    v2 = table.line_symbol(l2)
    v3 = table.line_symbol(l3)
    v4 = table.line_symbol(l4)
    expression = cast(sp.Expr, v1 - v2 - v3 + v4)  # type: ignore
    return [expression]


def enumerate_eqangles(
    table: AnglesTable, symbols: SymbolsRegistry
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for v1, v2, v3, v4 in table.expected_eqangles.copy():
        eqangle_expression: sp.Expr = cast(sp.Expr, v1 - v2 - v3 + v4)  # pyright: ignore
        if not table.inner_table.expr_delta(eqangle_expression):
            continue

        l1 = table.sympy_to_newclid_symbol[v1]
        l1_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l1)

        l2 = table.sympy_to_newclid_symbol[v2]
        l2_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l2)

        l3 = table.sympy_to_newclid_symbol[v3]
        l3_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l3)

        l4 = table.sympy_to_newclid_symbol[v4]
        l4_points = symbols.lines.get_all_points_on_line_and_equivalent_lines(l4)

        for pt_2, pt_3 in itertools.combinations(l1_points, 2):
            for pt_0, pt_1 in itertools.combinations(l2_points, 2):
                for pt_6, pt_7 in itertools.combinations(l3_points, 2):
                    for pt_4, pt_5 in itertools.combinations(l4_points, 2):
                        points = tuple(
                            PredicateArgument(pt.name)
                            for pt in (pt_0, pt_1, pt_2, pt_3, pt_4, pt_5, pt_6, pt_7)
                        )
                        eqangle_construction = (
                            PredicateConstruction.from_predicate_type_and_args(
                                PredicateType.EQUAL_ANGLES, points
                            )
                        )
                        yield eqangle_construction, eqangle_expression


def cong_expressions(
    segment1: Segment,
    segment2: Segment,
    table: RatiosTable,
) -> list[sp.Expr]:
    a, b = segment1
    v1 = table.segment_log_length((a, b))

    c, d = segment2
    v2 = table.segment_log_length((c, d))

    expression = cast(sp.Expr, v1 - v2)  # type: ignore
    return [expression]


def enumerate_congs(
    table: RatiosTable,
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for l1, l2 in table.expected_congs.copy():
        cong_expression = cast(sp.Expr, l1 - l2)  # pyright: ignore
        if not table.inner_table.expr_delta(cong_expression):
            continue

        p0, p1 = table.sympy_symbol_to_str_symbol[l1]
        p2, p3 = table.sympy_symbol_to_str_symbol[l2]

        args = tuple(PredicateArgument(pt.name) for pt in (p0, p1, p2, p3))
        congs_construction = PredicateConstruction.from_predicate_type_and_args(
            PredicateType.CONGRUENT, args
        )
        yield congs_construction, cong_expression


def eqratio_expressions(
    ratio1: tuple[Segment, Segment],
    ratio2: tuple[Segment, Segment],
    table: RatiosTable,
) -> list[sp.Expr]:
    (a, b), (c, d) = ratio1
    (e, f), (g, h) = ratio2
    v1 = table.segment_log_length((a, b))
    v2 = table.segment_log_length((c, d))
    v3 = table.segment_log_length((e, f))
    v4 = table.segment_log_length((g, h))
    expression = cast(sp.Expr, v1 - v2 - v3 + v4)  # type: ignore
    return [expression]


def enumerate_eqratios(
    table: RatiosTable,
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for l1, l2, l3, l4 in table.expected_eqratios.copy():
        eqratio_expression = cast(sp.Expr, l1 - l2 - l3 + l4)  # pyright: ignore
        if not table.inner_table.expr_delta(eqratio_expression):
            continue

        p0, p1 = table.sympy_symbol_to_str_symbol[l1]
        p2, p3 = table.sympy_symbol_to_str_symbol[l2]
        p4, p5 = table.sympy_symbol_to_str_symbol[l3]
        p6, p7 = table.sympy_symbol_to_str_symbol[l4]

        args = tuple(
            PredicateArgument(pt.name) for pt in (p0, p1, p2, p3, p4, p5, p6, p7)
        )
        eqratio_construction = PredicateConstruction.from_predicate_type_and_args(
            PredicateType.EQUAL_RATIOS, args
        )
        yield eqratio_construction, eqratio_expression


def constant_ratio_expressions(
    ratio: tuple[Segment, Segment],
    value: Fraction,
    table: RatiosTable,
) -> list[sp.Expr]:
    (a, b), (c, d) = ratio
    v1 = table.segment_log_length((a, b))
    v2 = table.segment_log_length((c, d))
    vr = sp.log(value.numerator) - sp.log(value.denominator)  # type: ignore

    expression = cast(sp.Expr, v1 - v2 - vr)  # type: ignore
    return [expression]


def enumerate_rconsts(
    table: RatiosTable,
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for (seg_1, seg_2), expected_ratio in table.expected_rconsts.copy().items():
        rconst_expression = cast(sp.Expr, seg_1 - seg_2)  # pyright: ignore
        subbed_in = table.inner_table.substitute_in_existing_expressions(
            rconst_expression
        )
        if not len(subbed_in.free_symbols) == 0:
            continue

        ratio: sp.Expr = sp.exp(subbed_in)  # pyright: ignore
        if not close_enough(float(ratio), expected_ratio):
            LOGGER.error(
                f"Ratio {ratio} is not equal to expected ratio {expected_ratio}"
            )

        n, d = sp.fraction(ratio)  # pyright: ignore

        p0, p1 = table.sympy_symbol_to_str_symbol[seg_1]
        p2, p3 = table.sympy_symbol_to_str_symbol[seg_2]
        fraction = PredicateArgument(fraction_to_ratio(Fraction(int(n), int(d))))
        points = tuple(PredicateArgument(pt.name) for pt in (p0, p1, p2, p3))
        rconst_construction = PredicateConstruction.from_predicate_type_and_args(
            PredicateType.CONSTANT_RATIO, points + (fraction,)
        )
        yield rconst_construction, rconst_expression


def constant_length_expression(
    segment: Segment,
    length: Fraction,
    table: RatiosTable,
) -> list[sp.Expr]:
    p0, p1 = segment
    line_var = table.segment_log_length((p0, p1))
    log_length = sp.log(length)
    expression = cast(sp.Expr, line_var - log_length)  # type: ignore
    return [expression]


def enumerate_lconsts(
    table: RatiosTable,
) -> Iterator[tuple[PredicateConstruction, sp.Expr]]:
    for segment, expected_length_value in table.expected_lconsts.copy():  # pyright: ignore
        lconst_expression = cast(sp.Expr, segment)  # pyright: ignore
        subbed_in = table.inner_table.substitute_in_existing_expressions(
            lconst_expression
        )
        if not len(subbed_in.free_symbols) == 0:
            continue

        length: sp.Expr = sp.exp(subbed_in)  # pyright: ignore
        if not close_enough(float(length), expected_length_value):  # pyright: ignore
            LOGGER.error(
                f"Length {length} is not equal to expected length {expected_length_value}"
            )

        p0, p1 = table.sympy_symbol_to_str_symbol[segment]
        points = tuple(PredicateArgument(pt.name) for pt in (p0, p1))

        length_arg = PredicateArgument(fraction_to_ratio(get_quotient(length)))
        lconst_construction = PredicateConstruction.from_predicate_type_and_args(
            PredicateType.CONSTANT_LENGTH, points + (length_arg,)
        )
        yield lconst_construction, lconst_expression



from __future__ import annotations

from typing import TYPE_CHECKING

from newclid.jgex.errors import PointTooCloseError, PointTooFarError
from newclid.jgex.sketch import LENGTH_UNIT
from newclid.predicate_types import PredicateArgument

if TYPE_CHECKING:
    from newclid.jgex.geometries import JGEXPoint


def ensure_not_too_close_numerical(
    newpoints: dict[PredicateArgument, JGEXPoint],
    existing_points: dict[PredicateArgument, JGEXPoint],
    tol: float = 0.01,
) -> bool:
    if len(existing_points) < 2:
        return False
    mindist = (
        sum(
            [
                sum([p.distance(p1) for p1 in existing_points.values() if p1 != p])
                for p in existing_points.values()
            ]
        )
        / len(existing_points)
        / (len(existing_points) - 1)
    )
    for new_point_name, new_point in newpoints.items():
        for existing_point_name, existing_point in existing_points.items():
            p2p_dist = new_point.distance(existing_point)
            if p2p_dist < tol * mindist:
                raise PointTooCloseError(
                    f"Point {new_point_name} is too close to {existing_point_name}"
                    f" (distance: {p2p_dist} < {tol * mindist})"
                )
    return False


def ensure_not_too_far_numerical(
    newpoints: dict[PredicateArgument, JGEXPoint],
    existing_points: dict[PredicateArgument, JGEXPoint],
    tol: float = 10.0,
) -> bool:
    if len(existing_points) < 2:
        return False
    maxdist = (
        sum(
            [
                sum([p.distance(p1) for p1 in existing_points.values() if p1 != p])
                for p in existing_points.values()
            ]
        )
        / len(existing_points)
        / (len(existing_points) - 1)
    )
    maxdist = max(maxdist, 3 * LENGTH_UNIT)
    for new_point_name, new_point in newpoints.items():
        for existing_point_name, existing_point in existing_points.items():
            p2p_dist = new_point.distance(existing_point)
            if p2p_dist > tol * maxdist:
                raise PointTooFarError(
                    f"Point {new_point_name} is too far from {existing_point_name}"
                    f" (distance: {p2p_dist} > {tol * maxdist})"
                )
    return False
